"""Initial migration

Revision ID: 9f2f70b48ab5
Revises:
Create Date: 2025-04-25 15:35:03.861966

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "9f2f70b48ab5"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "repos",
        sa.Column(
            "id", sa.UUID(), nullable=False, comment="MergeStat identifier for the repo"
        ),
        sa.Column("repo", sa.Text(), nullable=False, comment="URL for the repo"),
        sa.Column("ref", sa.Text(), nullable=True, comment="ref for the repo"),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            nullable=False,
            comment="timestamp of when the MergeStat repo entry was created",
        ),
        sa.Column(
            "settings", sa.JSON(), nullable=False, comment="JSON settings for the repo"
        ),
        sa.Column(
            "tags", sa.JSON(), nullable=False, comment="array of tags for the repo"
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "git_blame",
        sa.Column(
            "repo_id",
            sa.UUID(),
            nullable=False,
            comment="foreign key for public.repos.id",
        ),
        sa.Column(
            "path",
            sa.Text(),
            nullable=False,
            comment="path of the file the modification was made in",
        ),
        sa.Column(
            "line_no",
            sa.Integer(),
            nullable=False,
            comment="line number of the modification",
        ),
        sa.Column(
            "author_email",
            sa.Text(),
            nullable=True,
            comment="email of the author who modified the line",
        ),
        sa.Column(
            "author_name",
            sa.Text(),
            nullable=True,
            comment="name of the author who modified the line",
        ),
        sa.Column(
            "author_when",
            sa.DateTime(timezone=True),
            nullable=True,
            comment="timestamp of when the modification was authored",
        ),
        sa.Column(
            "commit_hash",
            sa.Text(),
            nullable=True,
            comment="hash of the commit the modification was made in",
        ),
        sa.Column("line", sa.Text(), nullable=True, comment="content of the line"),
        sa.Column(
            "_mergestat_synced_at",
            sa.DateTime(timezone=True),
            nullable=False,
            comment="timestamp when record was synced into the MergeStat database",
        ),
        sa.ForeignKeyConstraint(["repo_id"], ["repos.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("repo_id", "path", "line_no"),
    )
    op.create_table(
        "git_commit_stats",
        sa.Column(
            "repo_id",
            sa.UUID(),
            nullable=False,
            comment="foreign key for public.repos.id",
        ),
        sa.Column(
            "commit_hash", sa.Text(), nullable=False, comment="hash of the commit"
        ),
        sa.Column(
            "file_path",
            sa.Text(),
            nullable=False,
            comment="path of the file the modification was made in",
        ),
        sa.Column(
            "additions",
            sa.Integer(),
            nullable=False,
            comment="the number of additions in this path of the commit",
        ),
        sa.Column(
            "deletions",
            sa.Integer(),
            nullable=False,
            comment="the number of deletions in this path of the commit",
        ),
        sa.Column(
            "old_file_mode",
            sa.Text(),
            nullable=False,
            comment="old file mode derived from git mode",
        ),
        sa.Column(
            "new_file_mode",
            sa.Text(),
            nullable=True,
            comment="new file mode derived from git mode",
        ),
        sa.Column(
            "_mergestat_synced_at",
            sa.DateTime(timezone=True),
            nullable=False,
            comment="timestamp when record was synced into the MergeStat database",
        ),
        sa.ForeignKeyConstraint(["repo_id"], ["repos.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("repo_id", "commit_hash", "file_path"),
    )
    op.create_table(
        "git_commits",
        sa.Column(
            "repo_id",
            sa.UUID(),
            nullable=False,
            comment="foreign key for public.repos.id",
        ),
        sa.Column("hash", sa.Text(), nullable=False, comment="hash of the commit"),
        sa.Column("message", sa.Text(), nullable=True, comment="message of the commit"),
        sa.Column(
            "author_name",
            sa.Text(),
            nullable=True,
            comment="name of the author of the modification",
        ),
        sa.Column(
            "author_email",
            sa.Text(),
            nullable=True,
            comment="email of the author of the modification",
        ),
        sa.Column(
            "author_when",
            sa.DateTime(timezone=True),
            nullable=False,
            comment="timestamp of when the modification was authored",
        ),
        sa.Column(
            "committer_name",
            sa.Text(),
            nullable=True,
            comment="name of the author who committed the modification",
        ),
        sa.Column(
            "committer_email",
            sa.Text(),
            nullable=True,
            comment="email of the author who committed the modification",
        ),
        sa.Column(
            "committer_when",
            sa.DateTime(timezone=True),
            nullable=False,
            comment="timestamp of when the commit was made",
        ),
        sa.Column(
            "parents",
            sa.Integer(),
            nullable=False,
            comment="the number of parents of the commit",
        ),
        sa.Column(
            "_mergestat_synced_at",
            sa.DateTime(timezone=True),
            nullable=False,
            comment="timestamp when record was synced into the MergeStat database",
        ),
        sa.ForeignKeyConstraint(["repo_id"], ["repos.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("repo_id", "hash"),
    )
    op.create_table(
        "git_files",
        sa.Column(
            "repo_id",
            sa.UUID(),
            nullable=False,
            comment="foreign key for public.repos.id",
        ),
        sa.Column("path", sa.Text(), nullable=False, comment="path of the file"),
        sa.Column(
            "executable",
            sa.Boolean(),
            nullable=False,
            comment="boolean to determine if the file is an executable",
        ),
        sa.Column("contents", sa.Text(), nullable=True, comment="contents of the file"),
        sa.Column(
            "_mergestat_synced_at",
            sa.DateTime(timezone=True),
            nullable=False,
            comment="timestamp when record was synced into the MergeStat database",
        ),
        sa.ForeignKeyConstraint(["repo_id"], ["repos.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("repo_id", "path"),
    )
    op.create_table(
        "git_refs",
        sa.Column(
            "repo_id",
            sa.UUID(),
            nullable=False,
            comment="foreign key for public.repos.id",
        ),
        sa.Column("full_name", sa.Text(), nullable=False),
        sa.Column(
            "hash",
            sa.Text(),
            nullable=True,
            comment="hash of the commit for refs that are not of type tag",
        ),
        sa.Column("name", sa.Text(), nullable=True, comment="name of the ref"),
        sa.Column("remote", sa.Text(), nullable=True, comment="remote of the ref"),
        sa.Column("target", sa.Text(), nullable=True, comment="target of the ref"),
        sa.Column("type", sa.Text(), nullable=True, comment="type of the ref"),
        sa.Column(
            "tag_commit_hash",
            sa.Text(),
            nullable=True,
            comment="hash of the commit for refs that are of type tag",
        ),
        sa.Column(
            "_mergestat_synced_at",
            sa.DateTime(timezone=True),
            nullable=False,
            comment="timestamp when record was synced into the MergeStat database",
        ),
        sa.ForeignKeyConstraint(["repo_id"], ["repos.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("repo_id", "full_name"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("git_refs")
    op.drop_table("git_files")
    op.drop_table("git_commits")
    op.drop_table("git_commit_stats")
    op.drop_table("git_blame")
    op.drop_table("repos")
    # ### end Alembic commands ###



from __future__ import annotations

from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Sequence, Tuple
import logging

from models.work_items import WorkItem, WorkItemStatusTransition
from providers.identity import IdentityResolver
from providers.status_mapping import StatusMapping


def _to_utc(dt: Optional[datetime]) -> Optional[datetime]:
    if dt is None:
        return None
    if dt.tzinfo is None:
        return dt.replace(tzinfo=timezone.utc)
    return dt.astimezone(timezone.utc)


def _labels_from_nodes(nodes: Any) -> List[str]:
    labels: List[str] = []
    for node in nodes or []:
        name = (node or {}).get("name") if isinstance(node, dict) else getattr(node, "name", None)
        if name:
            labels.append(str(name))
    return labels


def github_issue_to_work_item(
    *,
    issue: Any,
    repo_full_name: str,
    repo_id: Optional[Any],
    status_mapping: StatusMapping,
    identity: IdentityResolver,
    events: Optional[Sequence[Any]] = None,
    project_status_raw: Optional[str] = None,
) -> Tuple[WorkItem, List[WorkItemStatusTransition]]:
    number = int(getattr(issue, "number", 0) or 0)
    work_item_id = f"gh:{repo_full_name}#{number}"

    title = getattr(issue, "title", "") or ""
    state = getattr(issue, "state", None)
    created_at = _to_utc(getattr(issue, "created_at", None)) or datetime.now(timezone.utc)
    updated_at = _to_utc(getattr(issue, "updated_at", None)) or created_at
    closed_at = _to_utc(getattr(issue, "closed_at", None))

    labels = [getattr(l, "name", None) for l in getattr(issue, "labels", []) or []]
    labels = [str(l) for l in labels if l]

    # If the issue is in a Project with a status field, prefer that as status_raw.
    status_raw = project_status_raw
    normalized_status = status_mapping.normalize_status(
        provider="github",
        status_raw=status_raw,
        labels=() if status_raw else labels,
        state=str(state) if state else None,
    )
    normalized_type = status_mapping.normalize_type(
        provider="github",
        type_raw=None,
        labels=labels,
    )

    assignees: List[str] = []
    for a in getattr(issue, "assignees", []) or []:
        assignees.append(
            identity.resolve(
                provider="github",
                email=getattr(a, "email", None),
                username=getattr(a, "login", None),
                display_name=getattr(a, "name", None),
            )
        )

    reporter_obj = getattr(issue, "user", None)
    reporter = None
    if reporter_obj is not None:
        reporter = identity.resolve(
            provider="github",
            email=getattr(reporter_obj, "email", None),
            username=getattr(reporter_obj, "login", None),
            display_name=getattr(reporter_obj, "name", None),
        )

    url = getattr(issue, "html_url", None) or getattr(issue, "url", None)

    # Best-effort transitions from issue events (label add/remove, closed/reopened).
    transitions: List[WorkItemStatusTransition] = []
    started_at = None
    completed_at = None
    if events:
        # Events are returned newest-first by PyGithub; sort by created_at.
        def _ev_dt(ev: Any) -> datetime:
            return _to_utc(getattr(ev, "created_at", None)) or datetime.min.replace(tzinfo=timezone.utc)

        prev_status = "unknown"
        for ev in sorted(list(events), key=_ev_dt):
            event_type = str(getattr(ev, "event", "") or "").lower()
            occurred_at = _to_utc(getattr(ev, "created_at", None)) or created_at

            if event_type in {"closed", "reopened"}:
                to_status = "done" if event_type == "closed" else "todo"
                transitions.append(
                    WorkItemStatusTransition(
                        work_item_id=work_item_id,
                        provider="github",
                        occurred_at=occurred_at,
                        from_status_raw=None,
                        to_status_raw=event_type,
                        from_status=prev_status,  # type: ignore[arg-type]
                        to_status=to_status,  # type: ignore[arg-type]
                        actor=None,
                    )
                )
                prev_status = to_status  # type: ignore[assignment]
                continue

            if event_type not in {"labeled", "unlabeled"}:
                continue

            label_obj = getattr(ev, "label", None)
            label_name = getattr(label_obj, "name", None)
            if not label_name:
                continue
            label_name = str(label_name)

            mapped = status_mapping.normalize_status(
                provider="github",
                status_raw=None,
                labels=[label_name] if event_type == "labeled" else (),
                state=None,
            )
            if mapped == "unknown":
                continue

            transitions.append(
                WorkItemStatusTransition(
                    work_item_id=work_item_id,
                    provider="github",
                    occurred_at=occurred_at,
                    from_status_raw=None,
                    to_status_raw=label_name,
                    from_status=prev_status,  # type: ignore[arg-type]
                    to_status=mapped,
                    actor=None,
                )
            )
            prev_status = mapped

        for t in transitions:
            if started_at is None and t.to_status == "in_progress":
                started_at = t.occurred_at
            if completed_at is None and t.to_status in {"done", "canceled"}:
                completed_at = t.occurred_at
                break

    # Fallback: closed_at implies done.
    if completed_at is None and closed_at is not None:
        completed_at = closed_at

    work_item = WorkItem(
        work_item_id=work_item_id,
        provider="github",
        repo_id=repo_id,
        project_key=None,
        # For work tracking metrics, treat the repo as the "project" scope.
        project_id=str(repo_full_name) if repo_full_name else None,
        title=str(title),
        type=normalized_type,
        status=normalized_status,
        status_raw=str(status_raw) if status_raw else (str(state) if state else None),
        assignees=[a for a in assignees if a and a != "unknown"],
        reporter=reporter if reporter and reporter != "unknown" else None,
        created_at=created_at,
        updated_at=updated_at,
        started_at=started_at,
        completed_at=completed_at,
        closed_at=closed_at,
        labels=labels,
        url=url,
    )

    return work_item, transitions


def github_project_v2_item_to_work_item(
    *,
    item_node: Dict[str, Any],
    project_scope_id: Optional[str] = None,
    status_mapping: StatusMapping,
    identity: IdentityResolver,
) -> Optional[WorkItem]:
    """
    Normalize a Projects v2 item node into a WorkItem.

    This is best-effort and intentionally does not attempt to reconstruct status history.
    """
    content = item_node.get("content") or {}
    typename = content.get("__typename")

    # Extract a status, iteration, and estimate values from field values.
    status_raw = None
    iteration_title = None
    iteration_id = None
    estimate = None
    
    for fv in (item_node.get("fieldValues") or {}).get("nodes") or []:
        typename = (fv or {}).get("__typename")
        field = (fv or {}).get("field") or {}
        field_name = str(field.get("name") or "").strip().lower()

        if typename == "ProjectV2ItemFieldSingleSelectValue":
            if field_name == "status":
                status_raw = fv.get("name")
        
        elif typename == "ProjectV2ItemFieldIterationValue":
            # GitHub Iterations
            if "iteration" in field_name or "sprint" in field_name:
                iteration_title = fv.get("title")
                iteration_id = fv.get("id") # internal node id
        
        elif typename == "ProjectV2ItemFieldNumberValue":
            # Estimates / Points
            if field_name in {"estimate", "points", "story points", "size"}:
                try:
                    estimate = float(fv.get("number") or 0)
                except (ValueError, TypeError):
                    logging.getLogger(__name__).debug(
                        "Failed to parse numeric estimate from value %r", fv.get("number")
                    )

    if typename == "Issue":
        repo_full_name = ((content.get("repository") or {}).get("nameWithOwner")) or ""
        number = int(content.get("number") or 0)
        work_item_id = f"gh:{repo_full_name}#{number}" if repo_full_name and number else f"ghproj:{item_node.get('id')}"
        labels = _labels_from_nodes(((content.get("labels") or {}).get("nodes")) or [])
        assignees = []
        for a in ((content.get("assignees") or {}).get("nodes")) or []:
            assignees.append(
                identity.resolve(
                    provider="github",
                    email=a.get("email"),
                    username=a.get("login"),
                    display_name=a.get("name"),
                )
            )
        author = content.get("author") or {}
        reporter = identity.resolve(
            provider="github",
            email=author.get("email"),
            username=author.get("login"),
            display_name=author.get("name"),
        )
        created_at = _to_utc(_parse_iso(content.get("createdAt"))) or datetime.now(timezone.utc)
        updated_at = _to_utc(_parse_iso(content.get("updatedAt"))) or created_at
        closed_at = _to_utc(_parse_iso(content.get("closedAt")))
        state = content.get("state")

        normalized_status = status_mapping.normalize_status(
            provider="github",
            status_raw=str(status_raw) if status_raw else None,
            labels=() if status_raw else labels,
            state=str(state) if state else None,
        )
        normalized_type = status_mapping.normalize_type(
            provider="github",
            type_raw=None,
            labels=labels,
        )
        completed_at = closed_at if closed_at else None

        return WorkItem(
            work_item_id=work_item_id,
            provider="github",
            repo_id=None,
            project_key=None,
            project_id=str(project_scope_id or repo_full_name) if (project_scope_id or repo_full_name) else None,
            title=str(content.get("title") or ""),
            type=normalized_type,
            status=normalized_status,
            status_raw=str(status_raw) if status_raw else (str(state) if state else None),
            assignees=[a for a in assignees if a and a != "unknown"],
            reporter=reporter if reporter and reporter != "unknown" else None,
            created_at=created_at,
            updated_at=updated_at,
            started_at=None,
            completed_at=completed_at,
            closed_at=closed_at,
            labels=labels,
            story_points=estimate,
            sprint_id=iteration_id,
            sprint_name=iteration_title,
            url=content.get("url"),
        )

    if typename == "DraftIssue":
        created_at = _to_utc(_parse_iso(content.get("createdAt"))) or datetime.now(timezone.utc)
        updated_at = _to_utc(_parse_iso(content.get("updatedAt"))) or created_at
        normalized_status = status_mapping.normalize_status(
            provider="github",
            status_raw=str(status_raw) if status_raw else None,
            labels=(),
            state=None,
        )
        return WorkItem(
            work_item_id=f"ghproj:{item_node.get('id')}",
            provider="github",
            repo_id=None,
            project_key=None,
            project_id=str(project_scope_id) if project_scope_id else None,
            title=str(content.get("title") or ""),
            type="issue",
            status=normalized_status,
            status_raw=str(status_raw) if status_raw else None,
            assignees=[],
            reporter=None,
            created_at=created_at,
            updated_at=updated_at,
            started_at=None,
            completed_at=None,
            closed_at=None,
            labels=[],
            story_points=estimate,
            sprint_id=iteration_id,
            sprint_name=iteration_title,
            url=None,
        )

    return None


def _parse_iso(value: Optional[str]) -> Optional[datetime]:
    if not value:
        return None
    try:
        # GitHub returns RFC3339 with Z.
        return datetime.fromisoformat(str(value).replace("Z", "+00:00"))
    except Exception:
        return None



from __future__ import annotations

from datetime import date, datetime, time, timedelta, timezone
from typing import Dict, List, Optional, Sequence, Tuple, TypedDict

from metrics.schemas import (
    WorkItemCycleTimeRecord,
    WorkItemMetricsDailyRecord,
    WorkItemUserMetricsDailyRecord,
)
from models.work_items import WorkItem, WorkItemStatusTransition
from providers.teams import TeamResolver
import logging


def _utc_day_window(day: date) -> Tuple[datetime, datetime]:
    start = datetime.combine(day, time.min, tzinfo=timezone.utc)
    end = start + timedelta(days=1)
    return start, end


def _to_utc(dt: datetime) -> datetime:
    if dt.tzinfo is None:
        return dt.replace(tzinfo=timezone.utc)
    return dt.astimezone(timezone.utc)


def _percentile(values: Sequence[float], percentile: float) -> float:
    if not values:
        return 0.0
    if percentile <= 0:
        return float(min(values))
    if percentile >= 100:
        return float(max(values))
    sorted_vals = sorted(float(v) for v in values)
    if len(sorted_vals) == 1:
        return float(sorted_vals[0])
    rank = (len(sorted_vals) - 1) * (float(percentile) / 100.0)
    lo = int(rank)
    hi = min(lo + 1, len(sorted_vals) - 1)
    frac = rank - lo
    return sorted_vals[lo] * (1.0 - frac) + sorted_vals[hi] * frac


def _resolve_team(
    team_resolver: Optional[TeamResolver], identity: Optional[str]
) -> Tuple[Optional[str], Optional[str]]:
    if team_resolver is None:
        return None, None
    return team_resolver.resolve(identity)


WAIT_STATUSES = {
    "backlog",
    "todo",
    "waiting",
    "blocked",
    "review_requested",
    "waiting_for_review",
}


def _calculate_flow_breakdown(
    item: WorkItem, transitions: List[WorkItemStatusTransition]
) -> Tuple[float, float]:
    if not item.started_at or not item.completed_at:
        return 0.0, 0.0

    start_utc = _to_utc(item.started_at)
    end_utc = _to_utc(item.completed_at)

    if start_utc >= end_utc:
        return 0.0, 0.0

    # Sort transitions by time
    sorted_trans = sorted(transitions, key=lambda x: x.occurred_at)
    
    # Filter transitions relevant to the cycle time window [started_at, completed_at]
    # Actually, we need to know the state *starting* from started_at.
    # We find the last transition *before* started_at to know initial state.
    
    # Simple approach: walk through time from start to end.
    current_status = "unknown"
    
    # Find initial status at started_at
    # Iterate backwards or keep track.
    # Assuming 'in_progress' is the start state if started_at is present.
    # But let's look at transitions before started_at.
    for t in sorted_trans:
        t_utc = _to_utc(t.occurred_at)
        if t_utc <= start_utc:
            current_status = t.to_status
        else:
            break
            
    # If explicitly started, and status is unknown or todo, maybe default to 'active' (in_progress)?
    # Usually started_at corresponds to a transition to In Progress.
    if current_status in ("unknown", "todo", "backlog"):
        current_status = "in_progress"

    active_seconds = 0.0
    wait_seconds = 0.0
    
    last_time = start_utc
    
    # Iterate transitions that happen *within* the window
    for t in sorted_trans:
        t_utc = _to_utc(t.occurred_at)
        if t_utc <= start_utc:
            continue
        if t_utc >= end_utc:
            break
            
        # Add duration of previous state
        duration = (t_utc - last_time).total_seconds()
        if current_status.lower() in WAIT_STATUSES:
            wait_seconds += duration
        else:
            active_seconds += duration
            
        # Update state and time
        current_status = t.to_status
        last_time = t_utc
        
    # Add final segment from last transition to completed_at
    duration = (end_utc - last_time).total_seconds()
    if duration > 0:
        if current_status.lower() in WAIT_STATUSES:
            wait_seconds += duration
        else:
            active_seconds += duration
            
    return active_seconds / 3600.0, wait_seconds / 3600.0


class GroupBucket(TypedDict):
    team_name: str
    items_started: int
    items_completed: int
    items_started_unassigned: int
    items_completed_unassigned: int
    wip_count: int
    wip_unassigned: int
    wip_age_hours: List[float]
    lead_hours: List[float]
    cycle_hours: List[float]
    bug_completed: int
    story_points_completed: float
    new_bugs: int
    new_items: int
    weekly_throughput: int
    predictability_score: float


class UserBucket(TypedDict):
    team_name: str
    items_started: int
    items_completed: int
    wip_count: int
    cycle_hours: List[float]


def compute_work_item_metrics_daily(
    *,
    day: date,
    work_items: Sequence[WorkItem],
    transitions: Sequence[WorkItemStatusTransition],
    computed_at: datetime,
    team_resolver: Optional[TeamResolver] = None,
) -> Tuple[
    List[WorkItemMetricsDailyRecord],
    List[WorkItemUserMetricsDailyRecord],
    List[WorkItemCycleTimeRecord],
]:
    """
    Compute work tracking metrics for a single UTC day.

    Inputs must be WorkItems with:
    - created_at, updated_at always set
    - started_at/completed_at best-effort derived (may be None)

    Null behavior:
    - cycle-time percentiles ignore items missing started_at or completed_at
    - WIP metrics ignore items missing started_at
    """
    start, end = _utc_day_window(day)
    computed_at_utc = _to_utc(computed_at)

    # Aggregations keyed by (provider, work_scope_id, team_id).
    by_group: Dict[Tuple[str, str, Optional[str]], GroupBucket] = {}
    by_user: Dict[Tuple[str, str, str, Optional[str]], UserBucket] = {}

    cycle_time_records: List[WorkItemCycleTimeRecord] = []

    # Pre-index transitions by work_item_id for faster lookup
    transitions_by_item: Dict[str, List[WorkItemStatusTransition]] = {}
    for t in transitions:
        transitions_by_item.setdefault(t.work_item_id, []).append(t)

    for item in work_items:
        work_scope_id = item.work_scope_id or ""
        created_at = _to_utc(item.created_at)
        started_at = _to_utc(item.started_at) if item.started_at else None
        completed_at = _to_utc(item.completed_at) if item.completed_at else None

        # Ignore items that don't exist yet on this day.
        if created_at >= end:
            continue

        assignee = item.assignees[0] if item.assignees else None
        team_id, team_name = _resolve_team(team_resolver, assignee)
        team_id_norm = team_id or "unassigned"
        team_name_norm = team_name or "Unassigned"

        started_today = started_at is not None and start <= started_at < end
        completed_today = completed_at is not None and start <= completed_at < end
        wip_end_of_day = started_at is not None and started_at < end and (completed_at is None or completed_at >= end)

        # Only emit a bucket for groups/users that have activity for this day.
        # However, for Phase 2 metrics (new items), we also need to account for items created today even if not started/completed.
        created_today = start <= created_at < end
        
        # We need to process if there's any activity or existence relevant to metrics
        relevant_activity = started_today or completed_today or wip_end_of_day or created_today
        if not relevant_activity:
            continue

        group_key = (item.provider, work_scope_id, team_id_norm)
        bucket = by_group.get(group_key)
        if bucket is None:
            bucket = {
                "team_name": team_name_norm,
                "items_started": 0,
                "items_completed": 0,
                "items_started_unassigned": 0,
                "items_completed_unassigned": 0,
                "wip_count": 0,
                "wip_unassigned": 0,
                "wip_age_hours": [],
                "lead_hours": [],
                "cycle_hours": [],
                "bug_completed": 0,
                "story_points_completed": 0.0,
                # Phase 2 metrics counters
                "new_bugs": 0,
                "new_items": 0,
                "weekly_throughput": 0,
                "predictability_score": 0.0,
            }
            by_group[group_key] = bucket

        user_identity = assignee or "unassigned"
        # User bucket (primary assignee or 'unassigned').
        if user_identity:
            user_key = (item.provider, work_scope_id, user_identity, team_id_norm)
            ub = by_user.get(user_key)
            if ub is None:
                ub = {
                    "team_name": team_name_norm,
                    "items_started": 0,
                    "items_completed": 0,
                    "wip_count": 0,
                    "cycle_hours": [],
                }
                by_user[user_key] = ub
        else:
            user_key = None
            ub = None

        # Phase 2: Creation stats
        if created_today:
            bucket["new_items"] += 1
            if item.type == "bug":
                bucket["new_bugs"] += 1

        # Phase 2: Weekly Throughput (Completed in last 7 days)
        # Window: [end - 7 days, end)
        week_start = end - timedelta(days=7)
        if completed_at is not None and week_start <= completed_at < end:
             bucket["weekly_throughput"] += 1

        # Started today.
        if started_today:
            bucket["items_started"] += 1
            if assignee is None:
                bucket["items_started_unassigned"] += 1
            if ub is not None:
                ub["items_started"] += 1

        # Completed today.
        if completed_today:
            # We already know completed_at is not None
            assert completed_at is not None
            bucket["items_completed"] += 1
            if assignee is None:
                bucket["items_completed_unassigned"] += 1
            if item.type == "bug":
                bucket["bug_completed"] += 1
            if item.story_points is not None:
                try:
                    bucket["story_points_completed"] += float(item.story_points)
                except Exception:
                    # Ignore invalid story_points values for this work item but log for diagnostics.
                    logging.getLogger(__name__).warning(
                        "Failed to convert story_points for work item %s: %r",
                        getattr(item, "work_item_id", None),
                        item.story_points,
                    )

            if ub is not None:
                ub["items_completed"] += 1

            lead_hours = (completed_at - created_at).total_seconds() / 3600.0
            bucket["lead_hours"].append(float(lead_hours))

            cycle_hours = None
            active_hours = None
            wait_hours = None
            flow_efficiency = None
            
            if started_at is not None:
                cycle_hours = (completed_at - started_at).total_seconds() / 3600.0
                bucket["cycle_hours"].append(float(cycle_hours))
                if ub is not None:
                    ub["cycle_hours"].append(float(cycle_hours))

            # Calculate flow breakdown if cycle_hours is available
            if cycle_hours is not None and cycle_hours > 0:
                item_transitions = transitions_by_item.get(item.work_item_id, [])
                calculated_active_h, calculated_wait_h = _calculate_flow_breakdown(item, item_transitions)
                
                # If no transitions recorded between start and complete, assume 100% active.
                if calculated_active_h + calculated_wait_h == 0:
                    active_hours = cycle_hours
                    wait_hours = 0.0
                else:
                    active_hours = calculated_active_h
                    wait_hours = calculated_wait_h
                
                flow_efficiency = (active_hours / (active_hours + wait_hours)) if (active_hours + wait_hours) > 0 else 0.0

            cycle_time_records.append(
                WorkItemCycleTimeRecord(
                    work_item_id=item.work_item_id,
                    provider=item.provider,
                    day=completed_at.date(), # type: ignore
                    work_scope_id=work_scope_id,
                    team_id=team_id or "unassigned",
                    team_name=team_name_norm,
                    assignee=assignee,
                    type=item.type,
                    status=item.status,
                    created_at=created_at,
                    started_at=started_at,
                    completed_at=completed_at,
                    cycle_time_hours=float(cycle_hours) if cycle_hours is not None else None,
                    lead_time_hours=float(lead_hours),
                    active_time_hours=float(active_hours) if active_hours is not None else None,
                    wait_time_hours=float(wait_hours) if wait_hours is not None else None,
                    flow_efficiency=float(flow_efficiency) if flow_efficiency is not None else None,
                    computed_at=computed_at_utc,
                )
            )

        # WIP (Active at end of day)
        if wip_end_of_day:
            # started_at is not None, checked by wip_end_of_day
            assert started_at is not None
            bucket["wip_count"] += 1
            if assignee is None:
                bucket["wip_unassigned"] += 1
            age_hours = (end - started_at).total_seconds() / 3600.0
            bucket["wip_age_hours"].append(float(age_hours))
            if ub is not None:
                ub["wip_count"] += 1

    group_records: List[WorkItemMetricsDailyRecord] = []
    for (provider, work_scope_id, team_id), bucket in sorted(
        by_group.items(), key=lambda kv: (kv[0][0], kv[0][1], str(kv[0][2] or ""))
    ):
        items_completed = bucket["items_completed"]
        bug_completed = bucket["bug_completed"]
        bug_ratio = (bug_completed / items_completed) if items_completed else 0.0
        cycle_hours = bucket["cycle_hours"]
        lead_hours = bucket["lead_hours"]
        wip_ages = bucket["wip_age_hours"]

        new_bugs = bucket["new_bugs"]
        new_items = bucket["new_items"]
        defect_rate = (new_bugs / new_items) if new_items else 0.0

        throughput_7d = bucket["weekly_throughput"]
        wip_val = bucket["wip_count"]
        # If throughput is 0, we can't divide. If WIP is > 0 and throughput is 0,
        # congestion is technically infinite. We'll cap or use 0.
        denominator = max(1.0, float(throughput_7d))
        wip_congestion = float(wip_val) / denominator

        # Predictability Proxy: Completion Rate (Completed / (Completed + Remaining))
        # This indicates how effectively the team clears its plate.
        total_load = float(items_completed + wip_val)
        predictability = (float(items_completed) / total_load) if total_load > 0 else 0.0
        
        group_records.append(
            WorkItemMetricsDailyRecord(
                day=day,
                provider=provider,
                work_scope_id=work_scope_id,
                team_id=team_id or "unassigned",
                team_name=bucket["team_name"],
                items_started=bucket["items_started"],
                items_completed=items_completed,
                items_started_unassigned=bucket["items_started_unassigned"],
                items_completed_unassigned=bucket["items_completed_unassigned"],
                wip_count_end_of_day=bucket["wip_count"],
                wip_unassigned_end_of_day=bucket["wip_unassigned"],
                cycle_time_p50_hours=float(_percentile(cycle_hours, 50.0)) if cycle_hours else None,
                cycle_time_p90_hours=float(_percentile(cycle_hours, 90.0)) if cycle_hours else None,
                lead_time_p50_hours=float(_percentile(lead_hours, 50.0)) if lead_hours else None,
                lead_time_p90_hours=float(_percentile(lead_hours, 90.0)) if lead_hours else None,
                wip_age_p50_hours=float(_percentile(wip_ages, 50.0)) if wip_ages else None,
                wip_age_p90_hours=float(_percentile(wip_ages, 90.0)) if wip_ages else None,
                bug_completed_ratio=float(bug_ratio),
                story_points_completed=float(bucket["story_points_completed"]),
                # Phase 2 metrics
                new_bugs_count=new_bugs,
                new_items_count=new_items,
                defect_intro_rate=defect_rate,
                wip_congestion_ratio=wip_congestion,
                predictability_score=predictability,
                computed_at=computed_at_utc,
            )
        )

    user_records: List[WorkItemUserMetricsDailyRecord] = []
    for (provider, work_scope_id, user_identity, team_id), bucket in sorted(
        by_user.items(), key=lambda kv: (kv[0][0], kv[0][1], kv[0][2], str(kv[0][3] or ""))
    ):
        cycle_hours = bucket["cycle_hours"]
        user_records.append(
            WorkItemUserMetricsDailyRecord(
                day=day,
                provider=provider,
                work_scope_id=work_scope_id,
                user_identity=user_identity,
                team_id=team_id or "unassigned",
                team_name=bucket["team_name"],
                items_started=bucket["items_started"],
                items_completed=bucket["items_completed"],
                wip_count_end_of_day=bucket["wip_count"],
                cycle_time_p50_hours=float(_percentile(cycle_hours, 50.0)) if cycle_hours else None,
                cycle_time_p90_hours=float(_percentile(cycle_hours, 90.0)) if cycle_hours else None,
                computed_at=computed_at_utc,
            )
        )

    return group_records, user_records, cycle_time_records



"""
Tests for batch repository processing features.
"""

from datetime import datetime, timezone
from unittest.mock import Mock, patch

import pytest

from connectors import GitHubConnector, BatchResult, match_repo_pattern
from models.git import GitCommit, GitCommitStat, get_repo_uuid_from_repo


@pytest.mark.asyncio
async def test_github_async_batch_callback_fires_as_completed(monkeypatch):
    """Fast repos should invoke callback before slow repos in same batch."""
    import time
    from unittest.mock import patch

    from connectors.models import Repository

    with (
        patch("connectors.github.Github"),
        patch("connectors.github.GitHubGraphQLClient"),
    ):
        connector = GitHubConnector(token="test_token")

    slow = Repository(
        id=1,
        name="slow",
        full_name="org/slow",
        default_branch="main",
        url="https://example.com/org/slow",
    )
    fast = Repository(
        id=2,
        name="fast",
        full_name="org/fast",
        default_branch="main",
        url="https://example.com/org/fast",
    )

    # Ensure input ordering doesn't accidentally match completion ordering.
    monkeypatch.setattr(
        connector,
        "_get_repositories_for_processing",
        lambda **kwargs: [slow, fast],
    )

    def fake_process(repo, max_commits):
        if repo.name == "slow":
            time.sleep(0.15)
        else:
            time.sleep(0.01)
        return BatchResult(repository=repo, stats=None, success=True)

    monkeypatch.setattr(connector, "_process_single_repo_stats", fake_process)

    callback_order = []

    await connector.get_repos_with_stats_async(
        org_name="org",
        batch_size=2,
        max_concurrent=2,
        rate_limit_delay=0,
        on_repo_complete=lambda r: callback_order.append(r.repository.name),
    )

    assert callback_order[0] == "fast"
    assert callback_order[1] == "slow"


@pytest.mark.asyncio
async def test_gitlab_async_batch_callback_fires_as_completed(monkeypatch):
    """Fast projects should invoke callback before slow projects in same batch."""
    import time
    from unittest.mock import patch

    from connectors.gitlab import GitLabBatchResult, GitLabConnector
    from connectors.models import Repository

    with patch("connectors.gitlab.gitlab.Gitlab"):
        connector = GitLabConnector(url="https://gitlab.com", private_token="test")

    slow = Repository(
        id=1,
        name="slow",
        full_name="group/slow",
        default_branch="main",
        url="https://example.com/group/slow",
    )
    fast = Repository(
        id=2,
        name="fast",
        full_name="group/fast",
        default_branch="main",
        url="https://example.com/group/fast",
    )

    monkeypatch.setattr(
        connector,
        "_get_projects_for_processing",
        lambda **kwargs: [slow, fast],
    )

    def fake_process(project, max_commits):
        if project.name == "slow":
            time.sleep(0.15)
        else:
            time.sleep(0.01)
        return GitLabBatchResult(project=project, stats=None, success=True)

    monkeypatch.setattr(connector, "_process_single_project_stats", fake_process)

    callback_order = []

    await connector.get_projects_with_stats_async(
        group_name="group",
        batch_size=2,
        max_concurrent=2,
        rate_limit_delay=0,
        on_project_complete=lambda r: callback_order.append(r.project.name),
    )

    assert callback_order[0] == "fast"
    assert callback_order[1] == "slow"


@pytest.mark.asyncio
async def test_process_gitlab_projects_batch_upserts_during_sync_processing(
    monkeypatch,
):
    """Default (sync) batch mode should still upsert before processing ends."""
    import threading
    import time

    import utils
    import processors.gitlab
    from connectors.gitlab import GitLabBatchResult

    # Force connectors availability for this test.
    monkeypatch.setattr(utils, "CONNECTORS_AVAILABLE", True)

    monkeypatch.setattr(
        processors.gitlab, "_fetch_gitlab_commits_sync", lambda *args, **kwargs: ([], [])
    )
    monkeypatch.setattr(
        processors.gitlab, "_fetch_gitlab_commit_stats_sync", lambda *args, **kwargs: []
    )

    inserted = threading.Event()

    class DummyStore:
        async def insert_repo(self, repo):
            inserted.set()

        async def insert_git_commit_data(self, commit_data):
            return

        async def insert_git_commit_stats(self, commit_stats):
            return

        async def insert_git_pull_requests(self, pr_data):
            return

    store = DummyStore()

    project = Mock()
    project.id = 123
    project.full_name = "group/proj"
    project.url = "https://example.com/group/proj"
    project.default_branch = "main"

    result = GitLabBatchResult(project=project, stats=None, success=True)

    class DummyConnector:
        def __init__(self, url: str, private_token: str):
            self.url = url
            self.private_token = private_token
            self.rest_client = Mock(get_merge_requests=lambda **kwargs: [])
            self.gitlab = Mock()
            self.gitlab.projects = Mock(get=lambda project_id: None)

        def get_projects_with_stats(self, **kwargs):
            on_project_complete = kwargs.get("on_project_complete")
            if on_project_complete:
                on_project_complete(result)

            # Give the event loop time to upsert while we're still running.
            deadline = time.time() + 2
            while time.time() < deadline and not inserted.is_set():
                time.sleep(0.01)

            assert inserted.is_set(), "Expected upsert during sync processing"
            return [result]

        def close(self):
            return

    monkeypatch.setattr(processors.gitlab, "GitLabConnector", DummyConnector)

    await processors.gitlab.process_gitlab_projects_batch(
        store=store,
        token="test_token",
        gitlab_url="https://gitlab.com",
        group_name="group",
        pattern="group/*",
        batch_size=1,
        max_concurrent=1,
        rate_limit_delay=0,
        use_async=False,
    )


@pytest.mark.asyncio
async def test_process_github_repos_batch_upserts_during_async_processing(monkeypatch):
    """Ensure async batch mode upserts as repos complete."""
    import asyncio

    import utils
    import processors.github

    # Force connectors availability for this test.
    monkeypatch.setattr(utils, "CONNECTORS_AVAILABLE", True)

    monkeypatch.setattr(
        processors.github, "_fetch_github_commits_sync", lambda *args, **kwargs: ([], [])
    )
    monkeypatch.setattr(
        processors.github, "_fetch_github_commit_stats_sync", lambda *args, **kwargs: []
    )

    # A store stub that records when insert_repo is called.
    inserted_event = asyncio.Event()

    class DummyStore:
        async def insert_repo(self, repo):
            inserted_event.set()

        async def insert_git_commit_data(self, commit_data):
            return

        async def insert_git_commit_stats(self, commit_stats):
            return

        async def insert_git_pull_requests(self, pr_data):
            return

    store = DummyStore()

    repo = Mock()
    repo.id = 123
    repo.full_name = "org/repo"
    repo.url = "https://example.com/org/repo"
    repo.default_branch = "main"
    repo.language = "Python"

    stats = Mock()
    stats.total_commits = 1
    stats.additions = 2
    stats.deletions = 1

    result = BatchResult(repository=repo, stats=stats, success=True)

    class DummyRepo:
        """Mock GitHub repository object."""

        def get_pulls(self, state="all"):
            """Return empty iterator for PRs."""
            return iter([])

    class DummyGithub:
        """Mock PyGithub Github object."""

        def get_repo(self, full_name: str):
            """Return a dummy repo."""
            return DummyRepo()

    class DummyConnector:
        def __init__(self, token: str):
            self.token = token
            self.github = DummyGithub()

        async def get_repos_with_stats_async(self, **kwargs):
            on_repo_complete = kwargs.get("on_repo_complete")
            if on_repo_complete:
                on_repo_complete(result)

            # Wait for the consumer to store the result.
            await asyncio.wait_for(inserted_event.wait(), timeout=2)
            return [result]

        def get_rate_limit(self):
            return {"remaining": 0, "limit": 0}

        def close(self):
            return

    monkeypatch.setattr(processors.github, "GitHubConnector", DummyConnector)

    await processors.github.process_github_repos_batch(
        store=store,
        token="test_token",
        org_name="org",
        pattern="org/*",
        batch_size=1,
        max_concurrent=1,
        rate_limit_delay=0,
        use_async=True,
    )


@pytest.mark.asyncio
async def test_process_github_repos_batch_stores_commits_and_stats(monkeypatch):
    """Batch GitHub processing should persist commits and stats for metrics."""
    import utils
    import processors.github

    # Force connectors availability and stub API helpers.
    monkeypatch.setattr(utils, "CONNECTORS_AVAILABLE", True)

    recorded_commits = []
    recorded_stats = []

    class DummyStore:
        async def insert_repo(self, repo):
            return

        async def insert_git_commit_data(self, commit_data):
            recorded_commits.extend(commit_data)

        async def insert_git_commit_stats(self, commit_stats):
            recorded_stats.extend(commit_stats)

        async def insert_git_pull_requests(self, pr_data):
            return

    repo = Mock()
    repo.id = 321
    repo.full_name = "org/repo-metrics"
    repo.url = "https://example.com/org/repo-metrics"
    repo.default_branch = "main"
    repo.language = "Python"

    stats = Mock()
    stats.total_commits = 1
    stats.additions = 2
    stats.deletions = 1

    result = BatchResult(repository=repo, stats=stats, success=True)

    def fake_fetch_commits(gh_repo, max_commits, repo_id, since=None):
        commit = GitCommit(
            repo_id=repo_id,
            hash="abc123",
            message="msg",
            author_name="alice",
            author_email=None,
            author_when=datetime.now(timezone.utc),
            committer_name="alice",
            committer_email=None,
            committer_when=datetime.now(timezone.utc),
            parents=1,
        )
        return ["raw"], [commit]

    def fake_fetch_commit_stats(raw_commits, repo_id, max_stats, since=None):
        return [
            GitCommitStat(
                repo_id=repo_id,
                commit_hash="abc123",
                file_path="file.txt",
                additions=1,
                deletions=0,
                old_file_mode="unknown",
                new_file_mode="unknown",
            )
        ]

    class DummyRepo:
        def get_pulls(self, state="all"):
            return iter([])

    class DummyGithub:
        def get_repo(self, full_name: str):
            return DummyRepo()

    class DummyConnector:
        def __init__(self, token: str):
            self.github = DummyGithub()

        async def get_repos_with_stats_async(self, **kwargs):
            on_repo_complete = kwargs.get("on_repo_complete")
            if on_repo_complete:
                on_repo_complete(result)
            return [result]

        def get_rate_limit(self):
            return {"remaining": 0, "limit": 0}

        def close(self):
            return

    monkeypatch.setattr(processors.github, "GitHubConnector", DummyConnector)
    monkeypatch.setattr(processors.github, "_fetch_github_commits_sync", fake_fetch_commits)
    monkeypatch.setattr(
        processors.github, "_fetch_github_commit_stats_sync", fake_fetch_commit_stats
    )

    store = DummyStore()

    await processors.github.process_github_repos_batch(
        store=store,
        token="test_token",
        org_name="org",
        pattern="org/*",
        batch_size=1,
        max_concurrent=1,
        rate_limit_delay=0,
        use_async=True,
    )

    assert any(c.hash == "abc123" for c in recorded_commits)
    expected_repo_id = get_repo_uuid_from_repo(repo.full_name)
    assert all(c.repo_id == expected_repo_id for c in recorded_commits)
    assert "abc123" in {s.commit_hash for s in recorded_stats}


@pytest.mark.asyncio
async def test_process_gitlab_projects_batch_stores_commits_and_stats(monkeypatch):
    """Batch GitLab processing should persist commits and stats for metrics."""
    import utils
    import processors.gitlab
    from connectors.gitlab import GitLabBatchResult

    monkeypatch.setattr(utils, "CONNECTORS_AVAILABLE", True)

    recorded_commits = []
    recorded_stats = []

    class DummyStore:
        async def insert_repo(self, repo):
            return

        async def insert_git_commit_data(self, commit_data):
            recorded_commits.extend(commit_data)

        async def insert_git_commit_stats(self, commit_stats):
            recorded_stats.extend(commit_stats)

        async def insert_git_pull_requests(self, pr_data):
            return

    project = Mock()
    project.id = 456
    project.full_name = "group/proj-metrics"
    project.url = "https://example.com/group/proj-metrics"
    project.default_branch = "main"

    result = GitLabBatchResult(project=project, stats=None, success=True)

    def fake_fetch_commits(gl_project, max_commits, repo_id, since=None):
        commit = GitCommit(
            repo_id=repo_id,
            hash="gitlab123",
            message="msg",
            author_name="bob",
            author_email=None,
            author_when=datetime.now(timezone.utc),
            committer_name="bob",
            committer_email=None,
            committer_when=datetime.now(timezone.utc),
            parents=1,
        )
        return ["raw"], [commit]

    def fake_fetch_commit_stats(gl_project, commit_hashes, repo_id, max_stats):
        return [
            GitCommitStat(
                repo_id=repo_id,
                commit_hash="gitlab123",
                file_path="file.txt",
                additions=2,
                deletions=1,
                old_file_mode="unknown",
                new_file_mode="unknown",
            )
        ]

    class DummyProjects:
        def get(self, project_id):
            return object()

    class DummyGitlab:
        def __init__(self):
            self.projects = DummyProjects()

    class DummyRestClient:
        def get_merge_requests(self, **kwargs):
            return []

    class DummyConnector:
        def __init__(self, url: str, private_token: str):
            self.url = url
            self.private_token = private_token
            self.gitlab = DummyGitlab()
            self.rest_client = DummyRestClient()

        async def get_projects_with_stats_async(self, **kwargs):
            on_project_complete = kwargs.get("on_project_complete")
            if on_project_complete:
                on_project_complete(result)
            return [result]

        def close(self):
            return

    monkeypatch.setattr(processors.gitlab, "GitLabConnector", DummyConnector)
    monkeypatch.setattr(processors.gitlab, "_fetch_gitlab_commits_sync", fake_fetch_commits)
    monkeypatch.setattr(
        processors.gitlab, "_fetch_gitlab_commit_stats_sync", fake_fetch_commit_stats
    )

    store = DummyStore()

    await processors.gitlab.process_gitlab_projects_batch(
        store=store,
        token="test_token",
        gitlab_url="https://gitlab.com",
        group_name="group",
        pattern="group/*",
        batch_size=1,
        max_concurrent=1,
        rate_limit_delay=0,
        use_async=True,
    )

    assert any(c.hash == "gitlab123" for c in recorded_commits)
    expected_repo_id = get_repo_uuid_from_repo(project.full_name)
    assert all(c.repo_id == expected_repo_id for c in recorded_commits)
    assert "gitlab123" in {s.commit_hash for s in recorded_stats}


class TestPatternMatching:
    """Test repository pattern matching functionality."""

    def test_exact_match(self):
        """Test exact repository name matching."""
        assert match_repo_pattern(
            "chrisgeo/mergestat-syncs", "chrisgeo/mergestat-syncs"
        )

    def test_wildcard_suffix(self):
        """Test pattern with wildcard suffix."""
        assert match_repo_pattern("chrisgeo/mergestat-syncs", "chrisgeo/merge*")
        assert match_repo_pattern("chrisgeo/mergestat", "chrisgeo/merge*")
        assert not match_repo_pattern("chrisgeo/other-repo", "chrisgeo/merge*")

    def test_wildcard_prefix(self):
        """Test pattern with wildcard prefix."""
        assert match_repo_pattern("chrisgeo/api-service", "*-service")
        assert match_repo_pattern("org/web-service", "*-service")

    def test_wildcard_owner(self):
        """Test pattern with wildcard owner."""
        assert match_repo_pattern("chrisgeo/sync-tool", "*/sync-tool")
        assert match_repo_pattern("otherorg/sync-tool", "*/sync-tool")

    def test_wildcard_repo(self):
        """Test pattern with wildcard repo."""
        assert match_repo_pattern("chrisgeo/anything", "chrisgeo/*")
        assert match_repo_pattern("chrisgeo/another", "chrisgeo/*")

    def test_case_insensitive(self):
        """Test case insensitive matching."""
        assert match_repo_pattern("ChrisGeo/MergeStat-Syncs", "chrisgeo/mergestat*")
        assert match_repo_pattern("CHRISGEO/REPO", "chrisgeo/*")

    def test_question_mark_wildcard(self):
        """Test question mark wildcard matching single character."""
        assert match_repo_pattern("chrisgeo/api-v1", "chrisgeo/api-v?")
        assert match_repo_pattern("chrisgeo/api-v2", "chrisgeo/api-v?")
        assert not match_repo_pattern("chrisgeo/api-v10", "chrisgeo/api-v?")

    def test_double_wildcard(self):
        """Test double wildcard matching."""
        assert match_repo_pattern("org/sub-api-service", "*api*")
        assert match_repo_pattern("chrisgeo/my-api", "*api*")

    def test_no_match(self):
        """Test non-matching patterns."""
        assert not match_repo_pattern("chrisgeo/repo", "other/*")
        assert not match_repo_pattern("org/api", "org/web*")


class TestBatchResult:
    """Test BatchResult dataclass."""

    def test_successful_result(self):
        """Test creating a successful batch result."""
        repo = Mock()
        repo.full_name = "org/repo"
        stats = Mock()

        result = BatchResult(repository=repo, stats=stats, success=True)

        assert result.repository == repo
        assert result.stats == stats
        assert result.success is True
        assert result.error is None

    def test_failed_result(self):
        """Test creating a failed batch result."""
        repo = Mock()
        repo.full_name = "org/repo"

        result = BatchResult(
            repository=repo,
            error="API error",
            success=False,
        )

        assert result.repository == repo
        assert result.stats is None
        assert result.success is False
        assert result.error == "API error"


class TestGitHubConnectorBatchProcessing:
    """Test GitHub connector batch processing features."""

    @pytest.fixture
    def mock_github_client(self):
        """Create a mock GitHub client."""
        with patch("connectors.github.Github") as mock_github:
            yield mock_github

    @pytest.fixture
    def mock_graphql_client(self):
        """Create a mock GraphQL client."""
        with patch("connectors.github.GitHubGraphQLClient") as mock_graphql:
            yield mock_graphql

    def _create_mock_repo(self, name: str, full_name: str):
        """Create a mock repository."""
        mock_repo = Mock()
        mock_repo.id = hash(full_name)
        mock_repo.name = name
        mock_repo.full_name = full_name
        mock_repo.default_branch = "main"
        mock_repo.description = f"Test repository {name}"
        mock_repo.html_url = f"https://github.com/{full_name}"
        mock_repo.created_at = None
        mock_repo.updated_at = None
        mock_repo.language = "Python"
        mock_repo.stargazers_count = 10
        mock_repo.forks_count = 5
        return mock_repo

    def test_list_repositories_with_pattern(
        self, mock_github_client, mock_graphql_client
    ):
        """Test listing repositories with pattern matching."""
        # Setup mock repos
        mock_repos = [
            self._create_mock_repo("mergestat-syncs", "chrisgeo/mergestat-syncs"),
            self._create_mock_repo("mergestat-lite", "chrisgeo/mergestat-lite"),
            self._create_mock_repo("other-repo", "chrisgeo/other-repo"),
            self._create_mock_repo("api-service", "chrisgeo/api-service"),
        ]

        mock_user = Mock()
        mock_user.get_repos.return_value = mock_repos

        mock_github_instance = mock_github_client.return_value
        mock_github_instance.get_user.return_value = mock_user

        # Test pattern matching using list_repositories with pattern parameter
        connector = GitHubConnector(token="test_token")
        repos = connector.list_repositories(
            user_name="chrisgeo",
            pattern="chrisgeo/merge*",
        )

        # Should only return repos matching the pattern
        assert len(repos) == 2
        assert all("merge" in repo.full_name.lower() for repo in repos)

    def test_list_repositories_with_pattern_max_repos(
        self, mock_github_client, mock_graphql_client
    ):
        """Test list_repositories with pattern respects max_repos."""
        # Setup mock repos
        mock_repos = [
            self._create_mock_repo(f"mergestat-{i}", f"chrisgeo/mergestat-{i}")
            for i in range(10)
        ]

        mock_user = Mock()
        mock_user.get_repos.return_value = mock_repos

        mock_github_instance = mock_github_client.return_value
        mock_github_instance.get_user.return_value = mock_user

        # Test with max_repos limit
        connector = GitHubConnector(token="test_token")
        repos = connector.list_repositories(
            user_name="chrisgeo",
            pattern="chrisgeo/merge*",
            max_repos=3,
        )

        assert len(repos) == 3

    def test_get_repos_with_stats_sync(self, mock_github_client, mock_graphql_client):
        """Test synchronous batch processing of repositories with stats."""
        # Setup mock repos
        mock_repos = [
            self._create_mock_repo("repo1", "chrisgeo/repo1"),
            self._create_mock_repo("repo2", "chrisgeo/repo2"),
        ]

        mock_user = Mock()
        mock_user.get_repos.return_value = mock_repos

        mock_github_instance = mock_github_client.return_value
        mock_github_instance.get_user.return_value = mock_user

        # Setup mock for get_repo (used by get_repo_stats)
        mock_gh_repo = Mock()
        mock_gh_repo.get_commits.return_value = []
        mock_gh_repo.created_at = None
        mock_github_instance.get_repo.return_value = mock_gh_repo

        # Test batch processing
        connector = GitHubConnector(token="test_token")
        results = connector.get_repos_with_stats(
            user_name="chrisgeo",
            batch_size=2,
            max_concurrent=2,
            rate_limit_delay=0.1,  # Small delay for testing
        )

        assert len(results) == 2
        assert all(isinstance(r, BatchResult) for r in results)

    def test_get_repos_with_stats_with_pattern(
        self, mock_github_client, mock_graphql_client
    ):
        """Test batch processing with pattern filtering."""
        # Setup mock repos
        mock_repos = [
            self._create_mock_repo("api-v1", "org/api-v1"),
            self._create_mock_repo("api-v2", "org/api-v2"),
            self._create_mock_repo("web-app", "org/web-app"),
        ]

        mock_org = Mock()
        mock_org.get_repos.return_value = mock_repos

        mock_github_instance = mock_github_client.return_value
        mock_github_instance.get_organization.return_value = mock_org

        # Setup mock for get_repo
        mock_gh_repo = Mock()
        mock_gh_repo.get_commits.return_value = []
        mock_gh_repo.created_at = None
        mock_github_instance.get_repo.return_value = mock_gh_repo

        # Test with pattern
        connector = GitHubConnector(token="test_token")
        results = connector.get_repos_with_stats(
            org_name="org",
            pattern="org/api-*",
            batch_size=5,
            rate_limit_delay=0.1,
        )

        # Should only process repos matching pattern
        assert len(results) == 2
        assert all("api-" in r.repository.full_name for r in results)

    def test_get_repos_with_stats_callback(
        self, mock_github_client, mock_graphql_client
    ):
        """Test callback is called for each processed repository."""
        # Setup mock repos
        mock_repos = [
            self._create_mock_repo("repo1", "user/repo1"),
            self._create_mock_repo("repo2", "user/repo2"),
        ]

        mock_user = Mock()
        mock_user.get_repos.return_value = mock_repos

        mock_github_instance = mock_github_client.return_value
        mock_github_instance.get_user.return_value = mock_user

        # Setup mock for get_repo
        mock_gh_repo = Mock()
        mock_gh_repo.get_commits.return_value = []
        mock_gh_repo.created_at = None
        mock_github_instance.get_repo.return_value = mock_gh_repo

        # Test callback
        callback = Mock()
        connector = GitHubConnector(token="test_token")
        _ = connector.get_repos_with_stats(
            user_name="user",
            on_repo_complete=callback,
            rate_limit_delay=0.1,
        )

        # Callback should be called for each repo
        assert callback.call_count == 2


class TestGitHubConnectorAsyncBatchProcessing:
    """Test GitHub connector async batch processing features."""

    @pytest.fixture
    def mock_github_client(self):
        """Create a mock GitHub client."""
        with patch("connectors.github.Github") as mock_github:
            yield mock_github

    @pytest.fixture
    def mock_graphql_client(self):
        """Create a mock GraphQL client."""
        with patch("connectors.github.GitHubGraphQLClient") as mock_graphql:
            yield mock_graphql

    def _create_mock_repo(self, name: str, full_name: str):
        """Create a mock repository."""
        mock_repo = Mock()
        mock_repo.id = hash(full_name)
        mock_repo.name = name
        mock_repo.full_name = full_name
        mock_repo.default_branch = "main"
        mock_repo.description = f"Test repository {name}"
        mock_repo.html_url = f"https://github.com/{full_name}"
        mock_repo.created_at = None
        mock_repo.updated_at = None
        mock_repo.language = "Python"
        mock_repo.stargazers_count = 10
        mock_repo.forks_count = 5
        return mock_repo

    @pytest.mark.asyncio
    async def test_get_repos_with_stats_async(
        self, mock_github_client, mock_graphql_client
    ):
        """Test async batch processing of repositories with stats."""
        # Setup mock repos
        mock_repos = [
            self._create_mock_repo("repo1", "chrisgeo/repo1"),
            self._create_mock_repo("repo2", "chrisgeo/repo2"),
        ]

        mock_user = Mock()
        mock_user.get_repos.return_value = mock_repos

        mock_github_instance = mock_github_client.return_value
        mock_github_instance.get_user.return_value = mock_user

        # Setup mock for get_repo
        mock_gh_repo = Mock()
        mock_gh_repo.get_commits.return_value = []
        mock_gh_repo.created_at = None
        mock_github_instance.get_repo.return_value = mock_gh_repo

        # Test async batch processing
        connector = GitHubConnector(token="test_token")
        results = await connector.get_repos_with_stats_async(
            user_name="chrisgeo",
            batch_size=2,
            max_concurrent=2,
            rate_limit_delay=0.1,
        )

        assert len(results) == 2
        assert all(isinstance(r, BatchResult) for r in results)

    @pytest.mark.asyncio
    async def test_get_repos_with_stats_async_with_pattern(
        self, mock_github_client, mock_graphql_client
    ):
        """Test async batch processing with pattern filtering."""
        # Setup mock repos
        mock_repos = [
            self._create_mock_repo("api-v1", "org/api-v1"),
            self._create_mock_repo("api-v2", "org/api-v2"),
            self._create_mock_repo("web-app", "org/web-app"),
        ]

        mock_org = Mock()
        mock_org.get_repos.return_value = mock_repos

        mock_github_instance = mock_github_client.return_value
        mock_github_instance.get_organization.return_value = mock_org

        # Setup mock for get_repo
        mock_gh_repo = Mock()
        mock_gh_repo.get_commits.return_value = []
        mock_gh_repo.created_at = None
        mock_github_instance.get_repo.return_value = mock_gh_repo

        # Test with pattern
        connector = GitHubConnector(token="test_token")
        results = await connector.get_repos_with_stats_async(
            org_name="org",
            pattern="org/api-*",
            batch_size=5,
            rate_limit_delay=0.1,
        )

        # Should only process repos matching pattern
        assert len(results) == 2
        assert all("api-" in r.repository.full_name for r in results)

    @pytest.mark.asyncio
    async def test_get_repos_with_stats_async_callback(
        self, mock_github_client, mock_graphql_client
    ):
        """Test async callback is called for each processed repository."""
        # Setup mock repos
        mock_repos = [
            self._create_mock_repo("repo1", "user/repo1"),
            self._create_mock_repo("repo2", "user/repo2"),
        ]

        mock_user = Mock()
        mock_user.get_repos.return_value = mock_repos

        mock_github_instance = mock_github_client.return_value
        mock_github_instance.get_user.return_value = mock_user

        # Setup mock for get_repo
        mock_gh_repo = Mock()
        mock_gh_repo.get_commits.return_value = []
        mock_gh_repo.created_at = None
        mock_github_instance.get_repo.return_value = mock_gh_repo

        # Test callback
        callback = Mock()
        connector = GitHubConnector(token="test_token")
        _ = await connector.get_repos_with_stats_async(
            user_name="user",
            on_repo_complete=callback,
            rate_limit_delay=0.1,
        )

        # Callback should be called for each repo
        assert callback.call_count == 2


class TestBatchProcessingErrorHandling:
    """Test error handling in batch repository processing."""

    @pytest.fixture
    def mock_github_client(self):
        """Create a mock GitHub client."""
        with patch("connectors.github.Github") as mock_github:
            yield mock_github

    @pytest.fixture
    def mock_graphql_client(self):
        """Create a mock GraphQL client."""
        with patch("connectors.github.GitHubGraphQLClient") as mock_graphql:
            yield mock_graphql

    def _create_mock_repo(self, name: str, full_name: str):
        """Create a mock repository."""
        mock_repo = Mock()
        mock_repo.id = hash(full_name)
        mock_repo.name = name
        mock_repo.full_name = full_name
        mock_repo.default_branch = "main"
        mock_repo.description = f"Test repository {name}"
        mock_repo.html_url = f"https://github.com/{full_name}"
        mock_repo.created_at = None
        mock_repo.updated_at = None
        mock_repo.language = "Python"
        mock_repo.stargazers_count = 10
        mock_repo.forks_count = 5
        return mock_repo

    def test_batch_processing_handles_api_error(
        self, mock_github_client, mock_graphql_client
    ):
        """Test batch processing continues when get_repo_stats raises an exception."""
        # Setup mock repos
        mock_repos = [
            self._create_mock_repo("repo1", "user/repo1"),
            self._create_mock_repo("repo2", "user/repo2"),
        ]

        mock_user = Mock()
        mock_user.get_repos.return_value = mock_repos

        mock_github_instance = mock_github_client.return_value
        mock_github_instance.get_user.return_value = mock_user

        # Setup mock for get_repo to raise an exception
        mock_github_instance.get_repo.side_effect = Exception("API rate limit exceeded")

        # Test batch processing
        connector = GitHubConnector(token="test_token")
        results = connector.get_repos_with_stats(
            user_name="user",
            batch_size=2,
            rate_limit_delay=0.1,
        )

        # All repos should be processed, but with errors
        assert len(results) == 2
        assert all(isinstance(r, BatchResult) for r in results)
        assert all(r.success is False for r in results)
        assert all(r.error is not None for r in results)
        assert all("API rate limit exceeded" in r.error for r in results)

    def test_batch_processing_handles_partial_failure(
        self, mock_github_client, mock_graphql_client
    ):
        """Test batch processing handles some repos failing while others succeed."""
        from datetime import datetime, timezone

        # Setup mock repos
        mock_repos = [
            self._create_mock_repo("repo1", "user/repo1"),
            self._create_mock_repo("repo2", "user/repo2"),
            self._create_mock_repo("repo3", "user/repo3"),
        ]

        mock_user = Mock()
        mock_user.get_repos.return_value = mock_repos

        mock_github_instance = mock_github_client.return_value
        mock_github_instance.get_user.return_value = mock_user

        # Setup mock for get_repo - first succeeds, second fails, third succeeds
        mock_gh_repo_success = Mock()
        mock_gh_repo_success.get_commits.return_value = []
        mock_gh_repo_success.created_at = datetime.now(timezone.utc)

        def get_repo_side_effect(repo_name):
            if "repo2" in repo_name:
                raise Exception("Repository not found")
            return mock_gh_repo_success

        mock_github_instance.get_repo.side_effect = get_repo_side_effect

        # Test batch processing
        connector = GitHubConnector(token="test_token")
        results = connector.get_repos_with_stats(
            user_name="user",
            batch_size=3,
            rate_limit_delay=0.1,
        )

        # All repos should be processed
        assert len(results) == 3

        # Count successes and failures
        successes = [r for r in results if r.success]
        failures = [r for r in results if not r.success]

        assert len(successes) == 2
        assert len(failures) == 1

        # Verify the failed repo has the correct error
        failed_result = failures[0]
        assert "Repository not found" in failed_result.error
        assert failed_result.repository.full_name == "user/repo2"

    def test_batch_processing_invalid_repo_name(
        self, mock_github_client, mock_graphql_client
    ):
        """Test batch processing handles invalid repository names."""
        # Setup mock repos with invalid name
        mock_repo_invalid = Mock()
        mock_repo_invalid.id = 1
        mock_repo_invalid.name = "invalid"
        mock_repo_invalid.full_name = (
            "invalid_no_slash"  # Invalid - no owner/repo format
        )
        mock_repo_invalid.default_branch = "main"
        mock_repo_invalid.description = "Invalid repo"
        mock_repo_invalid.html_url = "https://github.com/invalid"
        mock_repo_invalid.created_at = None
        mock_repo_invalid.updated_at = None
        mock_repo_invalid.language = "Python"
        mock_repo_invalid.stargazers_count = 0
        mock_repo_invalid.forks_count = 0

        mock_user = Mock()
        mock_user.get_repos.return_value = [mock_repo_invalid]

        mock_github_instance = mock_github_client.return_value
        mock_github_instance.get_user.return_value = mock_user

        # Test batch processing
        connector = GitHubConnector(token="test_token")
        results = connector.get_repos_with_stats(
            user_name="user",
            rate_limit_delay=0.1,
        )

        # Should have one result with error
        assert len(results) == 1
        assert results[0].success is False
        assert "Invalid repository name" in results[0].error

    @pytest.mark.asyncio
    async def test_async_batch_processing_handles_api_error(
        self, mock_github_client, mock_graphql_client
    ):
        """Test async batch processing continues when get_repo_stats raises an exception."""
        # Setup mock repos
        mock_repos = [
            self._create_mock_repo("repo1", "user/repo1"),
            self._create_mock_repo("repo2", "user/repo2"),
        ]

        mock_user = Mock()
        mock_user.get_repos.return_value = mock_repos

        mock_github_instance = mock_github_client.return_value
        mock_github_instance.get_user.return_value = mock_user

        # Setup mock for get_repo to raise an exception
        mock_github_instance.get_repo.side_effect = Exception("API rate limit exceeded")

        # Test async batch processing
        connector = GitHubConnector(token="test_token")
        results = await connector.get_repos_with_stats_async(
            user_name="user",
            batch_size=2,
            rate_limit_delay=0.1,
        )

        # All repos should be processed, but with errors
        assert len(results) == 2
        assert all(isinstance(r, BatchResult) for r in results)
        assert all(r.success is False for r in results)
        assert all(r.error is not None for r in results)


class TestGitLabPatternMatching:
    """Test GitLab project pattern matching functionality."""

    def test_exact_match(self):
        """Test exact project name matching."""
        from connectors import match_project_pattern

        assert match_project_pattern("group/project", "group/project")

    def test_wildcard_suffix(self):
        """Test pattern with wildcard suffix."""
        from connectors import match_project_pattern

        assert match_project_pattern("group/api-service", "group/api-*")
        assert match_project_pattern("group/api-v2", "group/api-*")
        assert not match_project_pattern("group/web-service", "group/api-*")

    def test_wildcard_prefix(self):
        """Test pattern with wildcard prefix."""
        from connectors import match_project_pattern

        assert match_project_pattern("mygroup/api-service", "*-service")
        assert match_project_pattern("other/web-service", "*-service")

    def test_wildcard_group(self):
        """Test pattern with wildcard group."""
        from connectors import match_project_pattern

        assert match_project_pattern("group1/sync-tool", "*/sync-tool")
        assert match_project_pattern("group2/sync-tool", "*/sync-tool")

    def test_wildcard_project(self):
        """Test pattern with wildcard project."""
        from connectors import match_project_pattern

        assert match_project_pattern("mygroup/anything", "mygroup/*")
        assert match_project_pattern("mygroup/another", "mygroup/*")

    def test_case_insensitive(self):
        """Test case insensitive matching."""
        from connectors import match_project_pattern

        assert match_project_pattern("MyGroup/MyProject", "mygroup/myproject*")
        assert match_project_pattern("MYGROUP/PROJECT", "mygroup/*")


class TestGitLabBatchResult:
    """Test GitLabBatchResult dataclass."""

    def test_successful_result(self):
        """Test creating a successful batch result."""
        from connectors import GitLabBatchResult

        project = Mock()
        project.full_name = "group/project"
        stats = Mock()

        result = GitLabBatchResult(project=project, stats=stats, success=True)

        assert result.project == project
        assert result.stats == stats
        assert result.success is True
        assert result.error is None

    def test_failed_result(self):
        """Test creating a failed batch result."""
        from connectors import GitLabBatchResult

        project = Mock()
        project.full_name = "group/project"

        result = GitLabBatchResult(
            project=project,
            error="API error",
            success=False,
        )

        assert result.project == project
        assert result.stats is None
        assert result.success is False
        assert result.error == "API error"


class TestGitLabConnectorBatchProcessing:
    """Test GitLab connector batch processing features."""

    @pytest.fixture
    def mock_gitlab_client(self):
        """Create a mock GitLab client."""
        with patch("connectors.gitlab.gitlab.Gitlab") as mock_gitlab:
            mock_instance = mock_gitlab.return_value
            mock_instance.auth.return_value = None
            yield mock_gitlab

    @pytest.fixture
    def mock_rest_client(self):
        """Create a mock REST client."""
        with patch("connectors.gitlab.GitLabRESTClient") as mock_rest:
            yield mock_rest

    def _create_mock_project(self, name: str, full_name: str):
        """Create a mock project."""
        mock_project = Mock()
        mock_project.id = hash(full_name)
        mock_project.name = name
        mock_project.path_with_namespace = full_name
        mock_project.default_branch = "main"
        mock_project.description = f"Test project {name}"
        mock_project.web_url = f"https://gitlab.com/{full_name}"
        mock_project.created_at = "2024-01-01T00:00:00Z"
        mock_project.last_activity_at = "2024-01-01T00:00:00Z"
        mock_project.star_count = 10
        mock_project.forks_count = 5
        return mock_project

    def test_list_projects_with_pattern(self, mock_gitlab_client, mock_rest_client):
        """Test listing projects with pattern matching."""
        from connectors import GitLabConnector

        # Setup mock projects
        mock_projects = [
            self._create_mock_project("api-service", "group/api-service"),
            self._create_mock_project("api-v2", "group/api-v2"),
            self._create_mock_project("web-app", "group/web-app"),
            self._create_mock_project("cli-tool", "group/cli-tool"),
        ]

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.projects.list.return_value = mock_projects

        # Test pattern matching
        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        projects = connector.list_projects(pattern="group/api-*")

        # Should only return projects matching the pattern
        assert len(projects) == 2
        assert all("api" in proj.full_name.lower() for proj in projects)

    def test_get_projects_with_stats_sync(self, mock_gitlab_client, mock_rest_client):
        """Test synchronous batch processing of projects with stats."""
        from connectors import GitLabConnector, GitLabBatchResult

        # Setup mock projects
        mock_projects = [
            self._create_mock_project("project1", "group/project1"),
            self._create_mock_project("project2", "group/project2"),
        ]

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.projects.list.return_value = mock_projects

        # Setup mock for get_repo_stats (via projects.get)
        mock_gl_project = Mock()
        mock_gl_project.commits.list.return_value = []
        mock_gl_project.created_at = "2024-01-01T00:00:00Z"
        mock_gitlab_instance.projects.get.return_value = mock_gl_project

        # Test batch processing
        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        results = connector.get_projects_with_stats(
            batch_size=2,
            max_concurrent=2,
            rate_limit_delay=0.1,  # Small delay for testing
        )

        assert len(results) == 2
        assert all(isinstance(r, GitLabBatchResult) for r in results)

    def test_get_projects_with_stats_with_pattern(
        self, mock_gitlab_client, mock_rest_client
    ):
        """Test batch processing with pattern filtering."""
        from connectors import GitLabConnector

        # Setup mock projects
        mock_projects = [
            self._create_mock_project("api-v1", "org/api-v1"),
            self._create_mock_project("api-v2", "org/api-v2"),
            self._create_mock_project("web-app", "org/web-app"),
        ]

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.projects.list.return_value = mock_projects

        # Setup mock group for pattern extraction (org/api-* -> group='org')
        mock_group = Mock()
        mock_group.projects.list.return_value = mock_projects
        mock_gitlab_instance.groups.get.return_value = mock_group

        # Setup mock for get_repo_stats
        mock_gl_project = Mock()
        mock_gl_project.commits.list.return_value = []
        mock_gl_project.created_at = "2024-01-01T00:00:00Z"
        mock_gitlab_instance.projects.get.return_value = mock_gl_project

        # Test with pattern
        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        results = connector.get_projects_with_stats(
            pattern="org/api-*",
            batch_size=5,
            rate_limit_delay=0.1,
        )

        # Should only process projects matching pattern
        assert len(results) == 2
        assert all("api-" in r.project.full_name for r in results)

    def test_get_projects_with_stats_callback(
        self, mock_gitlab_client, mock_rest_client
    ):
        """Test callback is called for each processed project."""
        from connectors import GitLabConnector

        # Setup mock projects
        mock_projects = [
            self._create_mock_project("project1", "group/project1"),
            self._create_mock_project("project2", "group/project2"),
        ]

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.projects.list.return_value = mock_projects

        # Setup mock for get_repo_stats
        mock_gl_project = Mock()
        mock_gl_project.commits.list.return_value = []
        mock_gl_project.created_at = "2024-01-01T00:00:00Z"
        mock_gitlab_instance.projects.get.return_value = mock_gl_project

        # Test callback
        callback = Mock()
        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        _ = connector.get_projects_with_stats(
            on_project_complete=callback,
            rate_limit_delay=0.1,
        )

        # Callback should be called for each project
        assert callback.call_count == 2


class TestGitLabConnectorAsyncBatchProcessing:
    """Test GitLab connector async batch processing features."""

    @pytest.fixture
    def mock_gitlab_client(self):
        """Create a mock GitLab client."""
        with patch("connectors.gitlab.gitlab.Gitlab") as mock_gitlab:
            mock_instance = mock_gitlab.return_value
            mock_instance.auth.return_value = None
            yield mock_gitlab

    @pytest.fixture
    def mock_rest_client(self):
        """Create a mock REST client."""
        with patch("connectors.gitlab.GitLabRESTClient") as mock_rest:
            yield mock_rest

    def _create_mock_project(self, name: str, full_name: str):
        """Create a mock project."""
        mock_project = Mock()
        mock_project.id = hash(full_name)
        mock_project.name = name
        mock_project.path_with_namespace = full_name
        mock_project.default_branch = "main"
        mock_project.description = f"Test project {name}"
        mock_project.web_url = f"https://gitlab.com/{full_name}"
        mock_project.created_at = "2024-01-01T00:00:00Z"
        mock_project.last_activity_at = "2024-01-01T00:00:00Z"
        mock_project.star_count = 10
        mock_project.forks_count = 5
        return mock_project

    @pytest.mark.asyncio
    async def test_get_projects_with_stats_async(
        self, mock_gitlab_client, mock_rest_client
    ):
        """Test async batch processing of projects with stats."""
        from connectors import GitLabConnector, GitLabBatchResult

        # Setup mock projects
        mock_projects = [
            self._create_mock_project("project1", "group/project1"),
            self._create_mock_project("project2", "group/project2"),
        ]

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.projects.list.return_value = mock_projects

        # Setup mock for get_repo_stats
        mock_gl_project = Mock()
        mock_gl_project.commits.list.return_value = []
        mock_gl_project.created_at = "2024-01-01T00:00:00Z"
        mock_gitlab_instance.projects.get.return_value = mock_gl_project

        # Test async batch processing
        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        results = await connector.get_projects_with_stats_async(
            batch_size=2,
            max_concurrent=2,
            rate_limit_delay=0.1,
        )

        assert len(results) == 2
        assert all(isinstance(r, GitLabBatchResult) for r in results)

    @pytest.mark.asyncio
    async def test_get_projects_with_stats_async_with_pattern(
        self, mock_gitlab_client, mock_rest_client
    ):
        """Test async batch processing with pattern filtering."""
        from connectors import GitLabConnector

        # Setup mock projects
        mock_projects = [
            self._create_mock_project("api-v1", "org/api-v1"),
            self._create_mock_project("api-v2", "org/api-v2"),
            self._create_mock_project("web-app", "org/web-app"),
        ]

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.projects.list.return_value = mock_projects

        # Setup mock group for pattern extraction (org/api-* -> group='org')
        mock_group = Mock()
        mock_group.projects.list.return_value = mock_projects
        mock_gitlab_instance.groups.get.return_value = mock_group

        # Setup mock for get_repo_stats
        mock_gl_project = Mock()
        mock_gl_project.commits.list.return_value = []
        mock_gl_project.created_at = "2024-01-01T00:00:00Z"
        mock_gitlab_instance.projects.get.return_value = mock_gl_project

        # Test with pattern
        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        results = await connector.get_projects_with_stats_async(
            pattern="org/api-*",
            batch_size=5,
            rate_limit_delay=0.1,
        )

        # Should only process projects matching pattern
        assert len(results) == 2
        assert all("api-" in r.project.full_name for r in results)

    @pytest.mark.asyncio
    async def test_get_projects_with_stats_async_callback(
        self, mock_gitlab_client, mock_rest_client
    ):
        """Test async callback is called for each processed project."""
        from connectors import GitLabConnector

        # Setup mock projects
        mock_projects = [
            self._create_mock_project("project1", "group/project1"),
            self._create_mock_project("project2", "group/project2"),
        ]

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.projects.list.return_value = mock_projects

        # Setup mock for get_repo_stats
        mock_gl_project = Mock()
        mock_gl_project.commits.list.return_value = []
        mock_gl_project.created_at = "2024-01-01T00:00:00Z"
        mock_gitlab_instance.projects.get.return_value = mock_gl_project

        # Test callback
        callback = Mock()
        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        _ = await connector.get_projects_with_stats_async(
            on_project_complete=callback,
            rate_limit_delay=0.1,
        )

        # Callback should be called for each project
        assert callback.call_count == 2


class TestGitLabBatchProcessingErrorHandling:
    """Test error handling in GitLab batch project processing."""

    @pytest.fixture
    def mock_gitlab_client(self):
        """Create a mock GitLab client."""
        with patch("connectors.gitlab.gitlab.Gitlab") as mock_gitlab:
            mock_instance = mock_gitlab.return_value
            mock_instance.auth.return_value = None
            yield mock_gitlab

    @pytest.fixture
    def mock_rest_client(self):
        """Create a mock REST client."""
        with patch("connectors.gitlab.GitLabRESTClient") as mock_rest:
            yield mock_rest

    def _create_mock_project(self, name: str, full_name: str):
        """Create a mock project."""
        mock_project = Mock()
        mock_project.id = hash(full_name)
        mock_project.name = name
        mock_project.path_with_namespace = full_name
        mock_project.default_branch = "main"
        mock_project.description = f"Test project {name}"
        mock_project.web_url = f"https://gitlab.com/{full_name}"
        mock_project.created_at = "2024-01-01T00:00:00Z"
        mock_project.last_activity_at = "2024-01-01T00:00:00Z"
        mock_project.star_count = 10
        mock_project.forks_count = 5
        return mock_project

    def test_batch_processing_handles_api_error(
        self, mock_gitlab_client, mock_rest_client
    ):
        """Test batch processing continues when get_repo_stats raises an exception."""
        from connectors import GitLabConnector, GitLabBatchResult

        # Setup mock projects
        mock_projects = [
            self._create_mock_project("project1", "group/project1"),
            self._create_mock_project("project2", "group/project2"),
        ]

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.projects.list.return_value = mock_projects

        # Setup mock for projects.get to raise an exception
        mock_gitlab_instance.projects.get.side_effect = Exception(
            "API rate limit exceeded"
        )

        # Test batch processing
        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        results = connector.get_projects_with_stats(
            batch_size=2,
            rate_limit_delay=0.1,
        )

        # All projects should be processed, but with errors
        assert len(results) == 2
        assert all(isinstance(r, GitLabBatchResult) for r in results)
        assert all(r.success is False for r in results)
        assert all(r.error is not None for r in results)

    def test_batch_processing_handles_partial_failure(
        self, mock_gitlab_client, mock_rest_client
    ):
        """Test batch processing handles some projects failing while others succeed."""
        from connectors import GitLabConnector

        # Setup mock projects
        mock_projects = [
            self._create_mock_project("project1", "group/project1"),
            self._create_mock_project("project2", "group/project2"),
            self._create_mock_project("project3", "group/project3"),
        ]

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.projects.list.return_value = mock_projects

        # Setup mock for projects.get - first succeeds, second fails, third succeeds
        mock_gl_project_success = Mock()
        mock_gl_project_success.commits.list.return_value = []
        mock_gl_project_success.created_at = "2024-01-01T00:00:00Z"

        def get_project_side_effect(project_name):
            if "project2" in project_name:
                raise Exception("Project not found")
            return mock_gl_project_success

        mock_gitlab_instance.projects.get.side_effect = get_project_side_effect

        # Test batch processing
        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        results = connector.get_projects_with_stats(
            batch_size=3,
            rate_limit_delay=0.1,
        )

        # All projects should be processed
        assert len(results) == 3

        # Count successes and failures
        successes = [r for r in results if r.success]
        failures = [r for r in results if not r.success]

        assert len(successes) == 2
        assert len(failures) == 1

        # Verify the failed project has the correct error
        failed_result = failures[0]
        assert "Project not found" in failed_result.error
        assert failed_result.project.full_name == "group/project2"

    @pytest.mark.asyncio
    async def test_async_batch_processing_handles_api_error(
        self, mock_gitlab_client, mock_rest_client
    ):
        """Test async batch processing continues when get_repo_stats raises an exception."""
        from connectors import GitLabConnector, GitLabBatchResult

        # Setup mock projects
        mock_projects = [
            self._create_mock_project("project1", "group/project1"),
            self._create_mock_project("project2", "group/project2"),
        ]

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.projects.list.return_value = mock_projects

        # Setup mock for projects.get to raise an exception
        mock_gitlab_instance.projects.get.side_effect = Exception(
            "API rate limit exceeded"
        )

        # Test async batch processing
        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        results = await connector.get_projects_with_stats_async(
            batch_size=2,
            rate_limit_delay=0.1,
        )

        # All projects should be processed, but with errors
        assert len(results) == 2
        assert all(isinstance(r, GitLabBatchResult) for r in results)
        assert all(r.success is False for r in results)
        assert all(r.error is not None for r in results)



#!/usr/bin/env python3
from __future__ import annotations

import argparse
import asyncio
import logging
import os
import subprocess
from datetime import date, datetime, time, timedelta, timezone
from pathlib import Path
from typing import List, Optional

from processors.github import process_github_repo, process_github_repos_batch
from processors.gitlab import process_gitlab_project, process_gitlab_projects_batch
from processors.local import process_local_blame, process_local_repo
from storage import create_store, detect_db_type
from utils import _parse_since

REPO_ROOT = Path(__file__).resolve().parent


def _load_dotenv(path: Path) -> int:
    """
    Load a .env file into process environment (without overriding existing vars).

    Keeps dependencies minimal (avoids python-dotenv).
    """
    if not path.exists():
        return 0
    loaded = 0
    for raw_line in path.read_text(encoding="utf-8").splitlines():
        line = raw_line.strip()
        if not line or line.startswith("#"):
            continue
        if line.startswith("export "):
            line = line[len("export ") :].strip()
        if "=" not in line:
            continue
        key, value = line.split("=", 1)
        key = key.strip()
        value = value.strip()
        if not key or key in os.environ:
            continue
        if (len(value) >= 2) and ((value[0] == value[-1]) and value[0] in {"'", '"'}):
            value = value[1:-1]
        os.environ[key] = value
        loaded += 1
    return loaded


def _parse_date(value: str) -> date:
    try:
        return date.fromisoformat(value)
    except ValueError as exc:
        raise argparse.ArgumentTypeError(
            f"Invalid date '{value}', expected YYYY-MM-DD"
        ) from exc


def _since_from_date_backfill(day: date, backfill_days: int) -> datetime:
    backfill = max(1, int(backfill_days))
    start_day = day - timedelta(days=backfill - 1)
    return datetime(
        start_day.year,
        start_day.month,
        start_day.day,
        tzinfo=timezone.utc,
    )


def _resolve_db_type(db_url: str, db_type: Optional[str]) -> str:
    if db_type:
        resolved = db_type.lower()
    else:
        try:
            resolved = detect_db_type(db_url)
        except ValueError as exc:
            raise SystemExit(str(exc)) from exc

    if resolved not in {"postgres", "mongo", "sqlite", "clickhouse"}:
        raise SystemExit(
            "DB_TYPE must be 'postgres', 'mongo', 'sqlite', or 'clickhouse'"
        )
    return resolved


async def _run_with_store(db_url: str, db_type: str, handler) -> None:
    store = create_store(db_url, db_type)
    async with store:
        await handler(store)


def _cmd_sync_teams(ns: argparse.Namespace) -> int:
    from models.teams import Team
    
    provider = (ns.provider or "config").lower()
    teams_data: List[Team] = []

    if provider == "config":
        from providers.teams import DEFAULT_TEAM_MAPPING_PATH
        import yaml

        path = Path(ns.path) if ns.path else DEFAULT_TEAM_MAPPING_PATH
        if not path.exists():
            logging.error(f"Teams config file not found at {path}")
            return 1

        try:
            with path.open("r", encoding="utf-8") as handle:
                payload = yaml.safe_load(handle) or {}
        except Exception as e:
            logging.error(f"Failed to parse teams config: {e}")
            return 1

        for entry in payload.get("teams") or []:
            team_id = str(entry.get("team_id") or "").strip()
            team_name = str(entry.get("team_name") or team_id).strip()
            description = entry.get("description")
            members = entry.get("members") or []
            if team_id:
                teams_data.append(
                    Team(
                        id=team_id,
                        name=team_name,
                        description=str(description) if description else None,
                        members=[str(m) for m in members],
                    )
                )

    elif provider == "jira":
        from providers.jira.client import JiraClient
        
        try:
            client = JiraClient.from_env()
        except ValueError as e:
            logging.error(f"Jira configuration error: {e}")
            return 1
            
        try:
            logging.info("Fetching projects from Jira...")
            projects = client.get_all_projects()
            for p in projects:
                # Use project Key as ID (stable), Name as Name
                key = p.get("key")
                name = p.get("name")
                desc = p.get("description")
                lead = p.get("lead", {})
                
                members = []
                if lead and lead.get("accountId"):
                    members.append(lead.get("accountId"))
                    
                if key and name:
                    teams_data.append(
                        Team(
                            id=key,
                            name=name,
                            description=str(desc) if desc else f"Jira Project {key}",
                            members=members
                        )
                    )
            logging.info(f"Fetched {len(teams_data)} projects from Jira.")
        except Exception as e:
            logging.error(f"Failed to fetch Jira projects: {e}")
            return 1
        finally:
            client.close()

    elif provider == "synthetic":
        from fixtures.generator import SyntheticDataGenerator
        generator = SyntheticDataGenerator()
        # Use 8 teams as requested for better visualization
        teams_data = generator.generate_teams(count=8)
        logging.info(f"Generated {len(teams_data)} synthetic teams.")

    else:
        logging.error(f"Unknown provider: {provider}")
        return 1

    if not teams_data:
        logging.warning("No teams found/generated.")
        return 0

    db_type = _resolve_db_type(ns.db, ns.db_type)

    async def _handler(store):
        # Ensure table exists (for SQL stores)
        if hasattr(store, "ensure_tables"):
            await store.ensure_tables()
        await store.insert_teams(teams_data)
        logging.info(f"Synced {len(teams_data)} teams to DB.")

    asyncio.run(_run_with_store(ns.db, db_type, _handler))
    return 0


def _resolve_since(ns: argparse.Namespace) -> Optional[datetime]:
    if ns.date is not None:
        return _since_from_date_backfill(ns.date, ns.backfill)
    if int(ns.backfill) != 1:
        raise SystemExit("--backfill requires --date")
    return _parse_since(getattr(ns, "since", None))


def _resolve_max_commits(ns: argparse.Namespace) -> Optional[int]:
    if ns.date is not None or getattr(ns, "since", None):
        return ns.max_commits_per_repo if ns.max_commits_per_repo is not None else None
    return ns.max_commits_per_repo or 100


def _sync_flags_for_target(target: str) -> dict:
    return {
        "sync_git": target == "git",
        "sync_prs": target == "prs",
        "sync_cicd": target == "cicd",
        "sync_deployments": target == "deployments",
        "sync_incidents": target == "incidents",
        "blame_only": target == "blame",
    }


def _resolve_synthetic_repo_name(ns: argparse.Namespace) -> str:
    if ns.repo_name:
        return ns.repo_name
    if ns.owner and ns.repo:
        return f"{ns.owner}/{ns.repo}"
    if ns.search:
        if "*" in ns.search or "?" in ns.search:
            raise SystemExit(
                "Synthetic provider does not support pattern search; use --repo-name."
            )
        return ns.search
    return "acme/demo-app"


def _cmd_sync_target(ns: argparse.Namespace) -> int:
    target = ns.sync_target
    provider = (ns.provider or "").lower()
    if provider not in {"local", "github", "gitlab", "synthetic"}:
        raise SystemExit("Provider must be one of: local, github, gitlab, synthetic.")

    if target not in {"git", "prs", "blame", "cicd", "deployments", "incidents"}:
        raise SystemExit("Sync target must be git, prs, blame, cicd, deployments, or incidents.")

    # TODO: Add Jira Ops/Service Desk incident ingestion once project-to-repo or deployment mapping is defined.
    if provider == "local":
        return _sync_local_target(ns, target)
    if provider == "github":
        return _sync_github_target(ns, target)
    if provider == "gitlab":
        return _sync_gitlab_target(ns, target)
    return _sync_synthetic_target(ns, target)


def _sync_local_target(ns: argparse.Namespace, target: str) -> int:
    if target not in {"git", "prs", "blame"}:
        raise SystemExit("Local provider supports only git, prs, or blame targets.")

    db_type = _resolve_db_type(ns.db, ns.db_type)
    since = _resolve_since(ns)

    async def _handler(store):
        if target == "blame":
            await process_local_blame(
                store=store,
                repo_path=ns.repo_path,
                since=since,
            )
            return

        await process_local_repo(
            store=store,
            repo_path=ns.repo_path,
            since=since,
            sync_git=(target == "git"),
            sync_prs=(target == "prs"),
            sync_blame=False,
        )

    asyncio.run(_run_with_store(ns.db, db_type, _handler))
    return 0


def _sync_github_target(ns: argparse.Namespace, target: str) -> int:
    token = ns.auth or os.getenv("GITHUB_TOKEN") or ""
    if not token:
        raise SystemExit("Missing GitHub token (pass --auth or set GITHUB_TOKEN).")

    db_type = _resolve_db_type(ns.db, ns.db_type)
    since = _resolve_since(ns)
    max_commits = _resolve_max_commits(ns)
    flags = _sync_flags_for_target(target)

    async def _handler(store):
        if ns.search:
            org_name = ns.group
            user_name = ns.owner if not ns.group else None
            await process_github_repos_batch(
                store=store,
                token=token,
                org_name=org_name,
                user_name=user_name,
                pattern=ns.search,
                batch_size=ns.batch_size,
                max_concurrent=ns.max_concurrent,
                rate_limit_delay=ns.rate_limit_delay,
                max_commits_per_repo=max_commits,
                max_repos=ns.max_repos,
                use_async=ns.use_async,
                sync_git=flags["sync_git"],
                sync_prs=flags["sync_prs"],
                sync_cicd=flags["sync_cicd"],
                sync_deployments=flags["sync_deployments"],
                sync_incidents=flags["sync_incidents"],
                blame_only=flags["blame_only"],
                backfill_missing=False,
                since=since,
            )
            return

        if not (ns.owner and ns.repo):
            raise SystemExit(
                "GitHub sync requires --owner and --repo (or --search for batch)."
            )
        await process_github_repo(
            store,
            ns.owner,
            ns.repo,
            token,
            blame_only=flags["blame_only"],
            max_commits=max_commits,
            sync_git=flags["sync_git"],
            sync_prs=flags["sync_prs"],
            sync_cicd=flags["sync_cicd"],
            sync_deployments=flags["sync_deployments"],
            sync_incidents=flags["sync_incidents"],
            since=since,
        )

    asyncio.run(_run_with_store(ns.db, db_type, _handler))
    return 0


def _sync_gitlab_target(ns: argparse.Namespace, target: str) -> int:
    token = ns.auth or os.getenv("GITLAB_TOKEN") or ""
    if not token:
        raise SystemExit("Missing GitLab token (pass --auth or set GITLAB_TOKEN).")

    db_type = _resolve_db_type(ns.db, ns.db_type)
    since = _resolve_since(ns)
    max_commits = _resolve_max_commits(ns)
    flags = _sync_flags_for_target(target)

    async def _handler(store):
        if ns.search:
            await process_gitlab_projects_batch(
                store=store,
                token=token,
                gitlab_url=ns.gitlab_url,
                group_name=ns.group,
                pattern=ns.search,
                batch_size=ns.batch_size,
                max_concurrent=ns.max_concurrent,
                rate_limit_delay=ns.rate_limit_delay,
                max_commits_per_project=max_commits,
                max_projects=ns.max_repos,
                use_async=ns.use_async,
                sync_git=flags["sync_git"],
                sync_prs=flags["sync_prs"],
                sync_cicd=flags["sync_cicd"],
                sync_deployments=flags["sync_deployments"],
                sync_incidents=flags["sync_incidents"],
                blame_only=flags["blame_only"],
                backfill_missing=False,
                since=since,
            )
            return

        if ns.project_id is None:
            raise SystemExit(
                "GitLab sync requires --project-id (or --search for batch)."
            )
        await process_gitlab_project(
            store,
            ns.project_id,
            token,
            ns.gitlab_url,
            blame_only=flags["blame_only"],
            max_commits=max_commits,
            sync_git=flags["sync_git"],
            sync_prs=flags["sync_prs"],
            sync_cicd=flags["sync_cicd"],
            sync_deployments=flags["sync_deployments"],
            sync_incidents=flags["sync_incidents"],
            since=since,
        )

    asyncio.run(_run_with_store(ns.db, db_type, _handler))
    return 0


def _sync_synthetic_target(ns: argparse.Namespace, target: str) -> int:
    from fixtures.generator import SyntheticDataGenerator

    repo_name = _resolve_synthetic_repo_name(ns)
    db_type = _resolve_db_type(ns.db, ns.db_type)
    days = max(1, int(ns.backfill))

    async def _handler(store):
        generator = SyntheticDataGenerator(repo_name=repo_name)
        repo = generator.generate_repo()
        await store.insert_repo(repo)

        if target == "git":
            commits = generator.generate_commits(days=days)
            await store.insert_git_commit_data(commits)
            stats = generator.generate_commit_stats(commits)
            await store.insert_git_commit_stats(stats)
            return

        if target == "prs":
            pr_data = generator.generate_prs()
            prs = [p["pr"] for p in pr_data]
            await store.insert_git_pull_requests(prs)

            reviews = []
            for p in pr_data:
                reviews.extend(p["reviews"])
            if reviews:
                await store.insert_git_pull_request_reviews(reviews)
            return

        if target == "blame":
            commits = generator.generate_commits(days=days)
            files = generator.generate_files()
            await store.insert_git_file_data(files)
            blame_data = generator.generate_blame(commits)
            if blame_data:
                await store.insert_blame_data(blame_data)
            return

        if target == "cicd":
            pipeline_runs = generator.generate_ci_pipeline_runs(days=days)
            if pipeline_runs:
                await store.insert_ci_pipeline_runs(pipeline_runs)
            return

        if target == "deployments":
            deployments = generator.generate_deployments(days=days)
            if deployments:
                await store.insert_deployments(deployments)
            return

        if target == "incidents":
            incidents = generator.generate_incidents(days=days)
            if incidents:
                await store.insert_incidents(incidents)
            return

    asyncio.run(_run_with_store(ns.db, db_type, _handler))
    return 0


def _cmd_sync_work_items(ns: argparse.Namespace) -> int:
    from metrics.job_work_items import run_work_items_sync_job

    if ns.auth:
        provider = (ns.provider or "").strip().lower()
        if provider == "gitlab":
            os.environ["GITLAB_TOKEN"] = ns.auth
        elif provider == "github":
            os.environ["GITHUB_TOKEN"] = ns.auth
        elif provider == "jira":
            raise SystemExit(
                "--auth is not supported for Jira; use JIRA_BASE_URL/JIRA_EMAIL/JIRA_API_TOKEN."
            )
        elif provider in {"all", "*"}:
            raise SystemExit(
                "--auth is ambiguous with --provider all; set GITHUB_TOKEN and GITLAB_TOKEN env vars."
            )

    run_work_items_sync_job(
        db_url=ns.db,
        day=ns.date,
        backfill_days=max(1, int(ns.backfill)),
        provider=ns.provider,
        sink=ns.sink,
        repo_id=ns.repo_id,
        repo_name=getattr(ns, "repo_name", None),
        search_pattern=getattr(ns, "search", None),
    )
    return 0


def _cmd_metrics_daily(ns: argparse.Namespace) -> int:
    # Import lazily to keep CLI startup fast and avoid optional deps at import time.
    from metrics.job_daily import run_daily_metrics_job

    run_daily_metrics_job(
        db_url=ns.db,
        day=ns.date,
        backfill_days=max(1, int(ns.backfill)),
        repo_id=ns.repo_id,
        repo_name=getattr(ns, "repo_name", None),
        include_commit_metrics=not ns.skip_commit_metrics,
        sink=ns.sink,
        provider=ns.provider,
    )
    return 0


def _cmd_metrics_complexity(ns: argparse.Namespace) -> int:
    from metrics.job_complexity import run_complexity_scan_job

    root_path = Path(ns.repo_path).resolve()

    if ns.repo_id:
        # Explicit single repo mode
        run_complexity_scan_job(
            repo_path=root_path,
            repo_id=ns.repo_id,
            db_url=ns.db,
            date=ns.date,
            backfill_days=ns.backfill,
            ref=ns.ref,
        )
        return 0

    # Batch mode: Fetch repos from DB and try to find them locally
    async def fetch_repos():
        db_type = _resolve_db_type(ns.db, None)
        store = create_store(ns.db, db_type)
        async with store:
            return await store.get_all_repos()

    try:
        repos = asyncio.run(fetch_repos())
    except Exception as e:
        logging.error(f"Failed to fetch repos from DB: {e}")
        return 1

    if not repos:
        logging.warning("No repositories found in database.")
        return 0

    # Filter by search pattern if provided
    if ns.search:
        import fnmatch

        original_count = len(repos)
        repos = [r for r in repos if fnmatch.fnmatch(r.repo, ns.search)]
        logging.info(f"Filtered repos by '{ns.search}': {len(repos)}/{original_count}")

    logging.info(
        f"Found {len(repos)} repositories in DB. Checking local availability in {root_path}..."
    )

    # Initialize sink once for batch processing
    db_type = _resolve_db_type(ns.db, None)
    sink = None
    if db_type == "clickhouse":
        from metrics.sinks.clickhouse import ClickHouseMetricsSink

        sink = ClickHouseMetricsSink(ns.db)
    elif db_type == "sqlite":
        from metrics.sinks.sqlite import SQLiteMetricsSink
        from metrics.job_complexity import _normalize_sqlite_url

        sink = SQLiteMetricsSink(_normalize_sqlite_url(ns.db))
    elif db_type == "mongo":
        from metrics.sinks.mongo import MongoMetricsSink

        sink = MongoMetricsSink(ns.db)
    elif db_type == "postgres":
        from metrics.sinks.sqlite import SQLiteMetricsSink
        from metrics.job_complexity import _normalize_postgres_url

        sink = SQLiteMetricsSink(_normalize_postgres_url(ns.db))

    if sink:
        sink.ensure_tables()

    processed_count = 0
    try:
        for repo in repos:
            # Heuristics to find repo locally based on DB identifier (name/URL)
            candidates = []
            if repo.repo:
                # 1. Check if root_path itself matches (by basename)
                repo_basename = Path(
                    repo.repo
                ).name  # e.g., 'dev-health-ops' from 'chrisgeo/dev-health-ops'
                if root_path.name == repo_basename and (root_path / ".git").exists():
                    candidates.append(root_path)
                # 2. Exact match (e.g. 'owner/repo' or 'repo') relative to root
                candidates.append(root_path / repo.repo)
                # 3. Basename match (e.g. 'repo' from 'owner/repo') relative to root
                candidates.append(root_path / Path(repo.repo).name)

            found_path = None
            for cand in candidates:
                if cand.is_dir() and (cand / ".git").exists():
                    found_path = cand
                    break

            if found_path:
                logging.info(
                    f"Processing DB repo '{repo.repo}' ({repo.id}) at {found_path}"
                )
                try:
                    run_complexity_scan_job(
                        repo_path=found_path,
                        repo_id=repo.id,
                        db_url=ns.db,
                        date=ns.date,
                        backfill_days=ns.backfill,
                        ref=ns.ref,
                        sink=sink,  # Pass existing sink
                    )
                    processed_count += 1
                except Exception as e:
                    logging.error(f"Failed to process {repo.repo}: {e}")
            else:
                logging.debug(f"Skipping DB repo '{repo.repo}': not found locally.")
    finally:
        if sink:
            sink.close()

    logging.info(f"Processed {processed_count} repositories.")
    return 0


def _cmd_grafana_up(_ns: argparse.Namespace) -> int:
    cmd = [
        "docker",
        "compose",
        "-f",
        str(REPO_ROOT / "compose.yml"),
        "up",
        "-d",
    ]
    return subprocess.run(cmd, check=False).returncode


def _cmd_grafana_down(_ns: argparse.Namespace) -> int:
    cmd = [
        "docker",
        "compose",
        "-f",
        str(REPO_ROOT / "compose.yml"),
        "down",
    ]
    return subprocess.run(cmd, check=False).returncode


def _add_sync_target_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("--db", required=True, help="Database connection string.")
    parser.add_argument(
        "--db-type",
        choices=["postgres", "mongo", "sqlite", "clickhouse"],
        help="Optional DB backend override.",
    )
    parser.add_argument(
        "--provider",
        choices=["local", "github", "gitlab", "synthetic"],
        required=True,
        help="Source provider for the sync job.",
    )
    parser.add_argument("--auth", help="Provider token override (GitHub/GitLab).")
    parser.add_argument(
        "--repo-path", default=".", help="Local git repo path (local provider)."
    )
    parser.add_argument("--owner", help="GitHub owner/org (single repo mode).")
    parser.add_argument("--repo", help="GitHub repo name (single repo mode).")
    parser.add_argument(
        "--project-id", type=int, help="GitLab project ID (single project mode)."
    )
    parser.add_argument(
        "--gitlab-url",
        default=os.getenv("GITLAB_URL", "https://gitlab.com"),
        help="GitLab instance URL.",
    )
    parser.add_argument("--group", help="Batch mode org/group name.")
    parser.add_argument(
        "-s",
        "--search",
        help="Batch mode pattern (e.g. 'org/*').",
    )
    parser.add_argument("--batch-size", type=int, default=10)
    parser.add_argument("--max-concurrent", type=int, default=4)
    parser.add_argument("--rate-limit-delay", type=float, default=1.0)
    parser.add_argument("--max-repos", type=int)
    parser.add_argument("--use-async", action="store_true")
    parser.add_argument("--max-commits-per-repo", type=int)
    parser.add_argument(
        "--repo-name", help="Synthetic repo name (default: acme/demo-app)."
    )
    time_group = parser.add_mutually_exclusive_group()
    time_group.add_argument("--since", help="Lower-bound ISO date/time (UTC).")
    time_group.add_argument(
        "--date",
        type=_parse_date,
        help="Target day (UTC) as YYYY-MM-DD (use with --backfill).",
    )
    parser.add_argument(
        "--backfill",
        type=int,
        default=1,
        help="Sync N days ending at --date (inclusive). Requires --date.",
    )


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="dev-health-ops",
        description="Sync git data and compute developer health metrics.",
    )
    parser.add_argument(
        "--log-level",
        default=os.getenv("LOG_LEVEL", "INFO"),
        help="Logging level (DEBUG, INFO, WARNING). Defaults to env LOG_LEVEL or INFO.",
    )
    sub = parser.add_subparsers(dest="command", required=True)

    # ---- sync ----
    # TODO: ship an installed binary (e.g. `dho sync ...`) instead of `python cli.py`.
    sync = sub.add_parser("sync", help="Sync git facts into a DB backend.")
    sync_sub = sync.add_subparsers(dest="sync_command", required=True)

    target_parsers = {
        "git": "Sync commits and commit stats.",
        "prs": "Sync pull/merge requests.",
        "blame": "Sync blame data only.",
        "cicd": "Sync CI/CD runs and pipelines.",
        "deployments": "Sync deployments.",
        "incidents": "Sync incidents.",
    }

    for target, help_text in target_parsers.items():
        target_parser = sync_sub.add_parser(target, help=help_text)
        _add_sync_target_args(target_parser)
        target_parser.set_defaults(func=_cmd_sync_target, sync_target=target)

    teams = sync_sub.add_parser("teams", help="Sync teams from config/teams.yaml, Jira, or Synthetic.")
    teams.add_argument("--db", required=True, help="Database connection string.")
    teams.add_argument(
        "--db-type",
        choices=["postgres", "mongo", "sqlite", "clickhouse"],
        help="Optional DB backend override.",
    )
    teams.add_argument(
        "--provider",
        choices=["config", "jira", "synthetic"],
        default="config",
        help="Source of team data (default: config).",
    )
    teams.add_argument("--path", help="Path to teams.yaml config (used if provider=config).")
    teams.set_defaults(func=_cmd_sync_teams)

    wi = sync_sub.add_parser(
        "work-items",
        help="Sync work tracking data from provider APIs (and write derived work item tables).",
    )
    wi.add_argument("--db", required=True, help="Database connection string.")
    wi.add_argument(
        "--provider",
        choices=["all", "jira", "github", "gitlab", "synthetic"],
        required=True,
        help="Which work item providers to sync.",
    )
    wi.add_argument(
        "--auth",
        help="Provider token override (GitHub/GitLab). Sets GITHUB_TOKEN or GITLAB_TOKEN.",
    )
    wi.add_argument(
        "--sink",
        choices=["auto", "clickhouse", "mongo", "sqlite", "postgres", "both"],
        default="auto",
        help="Where to write derived work item metrics.",
    )
    wi.add_argument(
        "--date",
        required=True,
        type=_parse_date,
        help="Target day (UTC) as YYYY-MM-DD.",
    )
    wi.add_argument(
        "--backfill",
        type=int,
        default=1,
        help="Sync and compute N days ending at --date (inclusive).",
    )
    wi.add_argument(
        "--repo-id",
        type=lambda s: __import__("uuid").UUID(s),
        help="Optional repo_id UUID filter (affects GitHub/GitLab repo selection).",
    )
    wi.add_argument(
        "--repo-name",
        help="Optional repo name filter (e.g. 'owner/repo') (affects GitHub/GitLab repo selection).",
    )
    wi.add_argument(
        "-s",
        "--search",
        help="Filter repos by name (glob pattern, e.g. 'org/*').",
    )
    wi.set_defaults(func=_cmd_sync_work_items)

    # ---- metrics ----
    metrics = sub.add_parser("metrics", help="Compute and write derived metrics.")
    metrics_sub = metrics.add_subparsers(dest="metrics_command", required=True)

    daily = metrics_sub.add_parser(
        "daily", help="Compute daily metrics (optionally backfill)."
    )
    daily.add_argument(
        "--date",
        required=True,
        type=_parse_date,
        help="Target day (UTC) as YYYY-MM-DD.",
    )
    daily.add_argument(
        "--backfill",
        type=int,
        default=1,
        help="Compute N days ending at --date (inclusive).",
    )
    daily.add_argument(
        "--db",
        default=os.getenv("DB_CONN_STRING") or os.getenv("DATABASE_URL"),
        help="Source DB URI (and default sink).",
    )
    daily.add_argument(
        "--repo-id",
        type=lambda s: __import__("uuid").UUID(s),
        help="Optional repo_id UUID filter.",
    )
    daily.add_argument(
        "--repo-name",
        help="Optional repo name filter (e.g. 'owner/repo').",
    )
    daily.add_argument(
        "--provider",
        choices=["all", "jira", "github", "gitlab", "synthetic", "none"],
        default="none",
        help="Which work item providers to include.",
    )
    daily.add_argument(
        "--sink",
        choices=["auto", "clickhouse", "mongo", "sqlite", "postgres", "both"],
        default="auto",
        help="Where to write derived metrics.",
    )
    daily.add_argument(
        "--skip-commit-metrics",
        action="store_true",
        help="Skip per-commit metrics output.",
    )
    daily.set_defaults(func=_cmd_metrics_daily)

    complexity = metrics_sub.add_parser(
        "complexity", help="Scan and compute complexity metrics."
    )
    complexity.add_argument(
        "--repo-path",
        default=".",
        help="Path to local git repo (or root dir for batch mode). Defaults to current dir.",
    )
    complexity.add_argument(
        "--repo-id",
        type=lambda s: __import__("uuid").UUID(s),
        help="Repo UUID. If omitted, scans dir for repos.",
    )
    complexity.add_argument(
        "-s", "--search", help="Filter repos by name (glob pattern, e.g. 'org/*')."
    )
    complexity.add_argument(
        "--date",
        type=_parse_date,
        default=date.today().isoformat(),
        help="Date of snapshot (YYYY-MM-DD).",
    )
    complexity.add_argument("--ref", default="HEAD", help="Git ref/branch analyzed.")
    complexity.add_argument(
        "--backfill",
        type=int,
        default=1,
        help="Compute N days ending at --date (inclusive).",
    )
    complexity.add_argument(
        "--db",
        default=os.getenv("CLICKHOUSE_DSN") or os.getenv("DB_CONN_STRING"),
        help="ClickHouse DSN.",
    )
    complexity.set_defaults(func=_cmd_metrics_complexity)

    # ---- grafana ----
    graf = sub.add_parser(
        "grafana", help="Start/stop the Grafana + ClickHouse dev stack."
    )
    graf_sub = graf.add_subparsers(dest="grafana_command", required=True)
    graf_up = graf_sub.add_parser(
        "up", help="docker compose up -d for grafana/docker-compose.yml"
    )
    graf_up.set_defaults(func=_cmd_grafana_up)
    graf_down = graf_sub.add_parser(
        "down", help="docker compose down for grafana/docker-compose.yml"
    )
    graf_down.set_defaults(func=_cmd_grafana_down)

    # ---- fixtures ----
    fix = sub.add_parser("fixtures", help="Data simulation and fixtures.")
    fix_sub = fix.add_subparsers(dest="fixtures_command", required=True)
    fix_gen = fix_sub.add_parser("generate", help="Generate synthetic data.")
    fix_gen.add_argument(
        "--db",
        default=os.getenv("DB_CONN_STRING") or os.getenv("DATABASE_URL"),
        help="Target DB URI.",
    )
    fix_gen.add_argument(
        "--db-type", help="Explicit DB type (postgres, clickhouse, etc)."
    )
    fix_gen.add_argument("--repo-name", default="acme/demo-app", help="Repo name.")
    fix_gen.add_argument("--days", type=int, default=30, help="Number of days of data.")
    fix_gen.add_argument(
        "--commits-per-day", type=int, default=5, help="Avg commits per day."
    )
    fix_gen.add_argument("--pr-count", type=int, default=20, help="Total PRs.")
    fix_gen.add_argument(
        "--with-metrics", action="store_true", help="Also generate derived metrics."
    )
    fix_gen.set_defaults(func=_cmd_fixtures_generate)

    return parser


def _cmd_fixtures_generate(ns: argparse.Namespace) -> int:
    from fixtures.generator import SyntheticDataGenerator

    generator = SyntheticDataGenerator(repo_name=ns.repo_name)
    db_type = _resolve_db_type(ns.db, ns.db_type)

    async def _handler(store):
        # 1. Repo
        repo = generator.generate_repo()
        await store.insert_repo(repo)

        # 2. Files
        files = generator.generate_files()
        await store.insert_git_file_data(files)

        # 3. Commits & Stats
        commits = generator.generate_commits(
            days=ns.days, commits_per_day=ns.commits_per_day
        )
        await store.insert_git_commit_data(commits)
        stats = generator.generate_commit_stats(commits)
        await store.insert_git_commit_stats(stats)

        # 4. PRs & Reviews
        pr_data = generator.generate_prs(count=ns.pr_count)
        prs = [p["pr"] for p in pr_data]
        await store.insert_git_pull_requests(prs)

        all_reviews = []
        for p in pr_data:
            all_reviews.extend(p["reviews"])
        if all_reviews:
            await store.insert_git_pull_request_reviews(all_reviews)

        # 5. CI/CD + Deployments + Incidents
        pr_numbers = [pr.number for pr in prs]
        pipeline_runs = generator.generate_ci_pipeline_runs(days=ns.days)
        deployments = generator.generate_deployments(
            days=ns.days, pr_numbers=pr_numbers
        )
        incidents = generator.generate_incidents(days=ns.days)
        await store.insert_ci_pipeline_runs(pipeline_runs)
        await store.insert_deployments(deployments)
        await store.insert_incidents(incidents)

        # 6. Blame Data (for Complexity)
        blame_data = generator.generate_blame(commits)
        if blame_data:
            await store.insert_blame_data(blame_data)

        logging.info(f"Generated synthetic data for {ns.repo_name}")
        logging.info(f"- Repo ID: {repo.id}")
        logging.info(f"- Commits: {len(commits)}")
        logging.info(f"- Blame lines: {len(blame_data)}")
        logging.info(f"- PRs: {len(prs)}")
        logging.info(f"- Reviews: {len(all_reviews)}")
        logging.info(f"- CI pipeline runs: {len(pipeline_runs)}")
        logging.info(f"- Deployments: {len(deployments)}")
        logging.info(f"- Incidents: {len(incidents)}")

        if ns.with_metrics:
            from metrics.job_daily import (
                ClickHouseMetricsSink,
                SQLiteMetricsSink,
            )

            sink = None
            if db_type == "clickhouse":
                sink = ClickHouseMetricsSink(ns.db)
            elif db_type == "sqlite":
                from metrics.job_daily import _normalize_sqlite_url

                sink = SQLiteMetricsSink(_normalize_sqlite_url(ns.db))

            if sink:
                sink.ensure_tables()
                from metrics.compute import compute_daily_metrics
                from metrics.compute_cicd import compute_cicd_metrics_daily
                from metrics.compute_deployments import compute_deploy_metrics_daily
                from metrics.compute_incidents import compute_incident_metrics_daily
                from metrics.compute_wellbeing import (
                    compute_team_wellbeing_metrics_daily,
                )
                from metrics.compute_work_item_state_durations import (
                    compute_work_item_state_durations_daily,
                )
                from metrics.compute_work_items import compute_work_item_metrics_daily
                from metrics.compute_ic import (
                    compute_ic_metrics_daily,
                    compute_ic_landscape_rolling,
                )
                from metrics.hotspots import compute_file_hotspots
                from metrics.quality import (
                    compute_rework_churn_ratio,
                    compute_single_owner_file_ratio,
                )
                from metrics.reviews import compute_review_edges_daily
                from providers.teams import TeamResolver

                commit_by_hash = {c.hash: c for c in commits}
                commit_stat_rows = []
                for stat in stats:
                    commit = commit_by_hash.get(stat.commit_hash)
                    if not commit:
                        continue
                    commit_stat_rows.append({
                        "repo_id": stat.repo_id,
                        "commit_hash": stat.commit_hash,
                        "author_email": commit.author_email,
                        "author_name": commit.author_name,
                        "committer_when": commit.committer_when,
                        "file_path": stat.file_path,
                        "additions": stat.additions,
                        "deletions": stat.deletions,
                    })

                pull_request_rows = []
                for pr in prs:
                    pull_request_rows.append({
                        "repo_id": pr.repo_id,
                        "number": pr.number,
                        "author_email": pr.author_email,
                        "author_name": pr.author_name,
                        "created_at": pr.created_at,
                        "merged_at": pr.merged_at,
                        "first_review_at": pr.first_review_at,
                        "first_comment_at": pr.first_comment_at,
                        "reviews_count": pr.reviews_count,
                        "comments_count": pr.comments_count,
                        "changes_requested_count": pr.changes_requested_count,
                        "additions": pr.additions,
                        "deletions": pr.deletions,
                        "changed_files": pr.changed_files,
                    })

                review_rows = []
                for review in all_reviews:
                    review_rows.append({
                        "repo_id": review.repo_id,
                        "number": review.number,
                        "reviewer": review.reviewer,
                        "submitted_at": review.submitted_at,
                        "state": review.state,
                    })

                pipeline_rows = []
                for run in pipeline_runs:
                    pipeline_rows.append({
                        "repo_id": run.repo_id,
                        "run_id": run.run_id,
                        "status": run.status,
                        "queued_at": run.queued_at,
                        "started_at": run.started_at,
                        "finished_at": run.finished_at,
                    })

                deployment_rows = []
                for deployment in deployments:
                    deployment_rows.append({
                        "repo_id": deployment.repo_id,
                        "deployment_id": deployment.deployment_id,
                        "status": deployment.status,
                        "environment": deployment.environment,
                        "started_at": deployment.started_at,
                        "finished_at": deployment.finished_at,
                        "deployed_at": deployment.deployed_at,
                        "merged_at": deployment.merged_at,
                        "pull_request_number": deployment.pull_request_number,
                    })

                incident_rows = []
                for incident in incidents:
                    incident_rows.append({
                        "repo_id": incident.repo_id,
                        "incident_id": incident.incident_id,
                        "status": incident.status,
                        "started_at": incident.started_at,
                        "resolved_at": incident.resolved_at,
                    })

                work_items = generator.generate_work_items(days=ns.days)
                transitions = generator.generate_work_item_transitions(work_items)

                # Provide stable team IDs for dashboards without requiring config.
                team_assignment = generator.get_team_assignment(count=8)
                team_resolver = TeamResolver(member_to_team=team_assignment["member_map"])

                computed_at = datetime.now(timezone.utc)

                # Generate and write complexity metrics
                if hasattr(sink, "write_file_complexity_snapshots"):
                    comp_data = generator.generate_complexity_metrics(days=ns.days)
                    if comp_data["snapshots"]:
                        sink.write_file_complexity_snapshots(comp_data["snapshots"])
                    if comp_data["dailies"]:
                        sink.write_repo_complexity_daily(comp_data["dailies"])
                    logging.info(f"- Complexity snapshots: {len(comp_data['snapshots'])}")

                end_day = computed_at.date()
                start_day = end_day - timedelta(days=max(1, int(ns.days)) - 1)

                total_wi = 0
                total_wi_users = 0
                total_cycles = 0
                total_state = 0
                total_repos = 0
                total_users = 0
                total_commits = 0
                total_files = 0
                total_teams = 0
                total_review_edges = 0
                total_cicd = 0
                total_deploy = 0
                total_incident = 0

                for i in range(max(1, int(ns.days))):
                    day = start_day + timedelta(days=i)

                    start_dt = datetime.combine(day, time.min, tzinfo=timezone.utc)
                    end_dt = start_dt + timedelta(days=1)

                    mttr_by_repo = {}
                    bug_times = {}
                    for item in work_items:
                        if item.type == "bug" and item.completed_at and item.started_at:
                            start_item = item.started_at
                            end_item = item.completed_at
                            if start_item is None or end_item is None:
                                continue
                            if start_item < end_dt and end_item >= start_dt:
                                r_id = getattr(item, "repo_id", None)
                                if r_id:
                                    hours = (
                                        end_item - start_item
                                    ).total_seconds() / 3600.0
                                    bug_times.setdefault(r_id, []).append(hours)
                    for r_id, times in bug_times.items():
                        mttr_by_repo[r_id] = sum(times) / len(times)

                    window_days = 30
                    h_start = datetime.combine(
                        day - timedelta(days=window_days - 1),
                        time.min,
                        tzinfo=timezone.utc,
                    )
                    window_stats = [
                        row
                        for row in commit_stat_rows
                        if h_start <= row["committer_when"] < end_dt
                    ]
                    day_commit_stats = [
                        row
                        for row in commit_stat_rows
                        if start_dt <= row["committer_when"] < end_dt
                    ]
                    rework_ratio_by_repo = {
                        repo.id: compute_rework_churn_ratio(
                            repo_id=str(repo.id), window_stats=window_stats
                        )
                    }
                    single_owner_ratio_by_repo = {
                        repo.id: compute_single_owner_file_ratio(
                            repo_id=str(repo.id), window_stats=window_stats
                        )
                    }

                    repo_result = compute_daily_metrics(
                        day=day,
                        commit_stat_rows=day_commit_stats,
                        pull_request_rows=pull_request_rows,
                        pull_request_review_rows=review_rows,
                        computed_at=computed_at,
                        include_commit_metrics=True,
                        team_resolver=team_resolver,
                        mttr_by_repo=mttr_by_repo,
                        rework_churn_ratio_by_repo=rework_ratio_by_repo,
                        single_owner_file_ratio_by_repo=single_owner_ratio_by_repo,
                    )
                    team_metrics = compute_team_wellbeing_metrics_daily(
                        day=day,
                        commit_stat_rows=day_commit_stats,
                        team_resolver=team_resolver,
                        computed_at=computed_at,
                    )

                    wi_rows, wi_user_rows, cycle_rows = compute_work_item_metrics_daily(
                        day=day,
                        work_items=work_items,
                        transitions=transitions,
                        computed_at=computed_at,
                        team_resolver=team_resolver,
                    )

                    # Enrich User Metrics with IC fields
                    # Convert TeamResolver map to simple identity->team_id map
                    team_map = {k: v[0] for k, v in team_assignment["member_map"].items()}
                    ic_metrics = compute_ic_metrics_daily(
                        git_metrics=repo_result.user_metrics,
                        wi_metrics=wi_user_rows,
                        team_map=team_map,
                    )
                    # Use the enriched list for writing
                    repo_result.user_metrics[:] = ic_metrics

                    if wi_rows:
                        sink.write_work_item_metrics(wi_rows)
                        total_wi += len(wi_rows)
                    if wi_user_rows:
                        sink.write_work_item_user_metrics(wi_user_rows)
                        total_wi_users += len(wi_user_rows)
                    if cycle_rows:
                        sink.write_work_item_cycle_times(cycle_rows)
                        total_cycles += len(cycle_rows)

                    state_rows = compute_work_item_state_durations_daily(
                        day=day,
                        work_items=work_items,
                        transitions=transitions,
                        computed_at=computed_at,
                        team_resolver=team_resolver,
                    )
                    if state_rows:
                        sink.write_work_item_state_durations(state_rows)
                        total_state += len(state_rows)

                    file_metrics = compute_file_hotspots(
                        repo_id=repo.id,
                        day=day,
                        window_stats=window_stats,
                        computed_at=computed_at,
                    )
                    if file_metrics:
                        sink.write_file_metrics(file_metrics)
                        total_files += len(file_metrics)

                    review_edges = compute_review_edges_daily(
                        day=day,
                        pull_request_rows=pull_request_rows,
                        pull_request_review_rows=review_rows,
                        computed_at=computed_at,
                    )
                    if review_edges:
                        sink.write_review_edges(review_edges)
                        total_review_edges += len(review_edges)

                    if repo_result.repo_metrics:
                        sink.write_repo_metrics(repo_result.repo_metrics)
                        total_repos += len(repo_result.repo_metrics)
                    if repo_result.user_metrics:
                        sink.write_user_metrics(repo_result.user_metrics)
                        total_users += len(repo_result.user_metrics)
                    if repo_result.commit_metrics:
                        sink.write_commit_metrics(repo_result.commit_metrics)
                        total_commits += len(repo_result.commit_metrics)
                    if team_metrics:
                        sink.write_team_metrics(team_metrics)
                        total_teams += len(team_metrics)

                    cicd_metrics = compute_cicd_metrics_daily(
                        day=day,
                        pipeline_runs=pipeline_rows,
                        computed_at=computed_at,
                    )
                    if cicd_metrics:
                        sink.write_cicd_metrics(cicd_metrics)
                        total_cicd += len(cicd_metrics)

                    deploy_metrics = compute_deploy_metrics_daily(
                        day=day,
                        deployments=deployment_rows,
                        computed_at=computed_at,
                    )
                    if deploy_metrics:
                        sink.write_deploy_metrics(deploy_metrics)
                        total_deploy += len(deploy_metrics)

                    incident_metrics = compute_incident_metrics_daily(
                        day=day,
                        incidents=incident_rows,
                        computed_at=computed_at,
                    )
                    if incident_metrics:
                        sink.write_incident_metrics(incident_metrics)
                        total_incident += len(incident_metrics)

                    # Landscape rolling metrics
                    try:
                        # For fixtures, we can't easily query the DB we just wrote to in the same transaction context/loop easily
                        # if using batch inserts or if it's not committed.
                        # However, for ClickHouse sink, inserts are immediate.
                        # For SQLite/Postgres, it depends on the driver/transaction.
                        # But sink methods handle their own transactions usually.
                        # Let's try to compute it.
                        if hasattr(sink, "get_rolling_30d_user_stats") and hasattr(
                            sink, "write_ic_landscape_rolling"
                        ):
                            rolling_stats = sink.get_rolling_30d_user_stats(
                                as_of_day=day, repo_id=repo.id
                            )
                            landscape_recs = compute_ic_landscape_rolling(
                                as_of_day=day,
                                rolling_stats=rolling_stats,
                                team_map=team_map,
                            )
                            sink.write_ic_landscape_rolling(landscape_recs)
                    except Exception as e:
                        logging.warning(
                            "Failed to compute/write fixture landscape metrics: %s", e
                        )

                logging.info(
                    "Generated fixtures metrics: repo=%d user=%d commit=%d team=%d file=%d review_edges=%d work_item_metrics=%d work_item_user_metrics=%d work_item_cycle_times=%d work_item_state_durations=%d cicd=%d deploy=%d incident=%d",
                    total_repos,
                    total_users,
                    total_commits,
                    total_teams,
                    total_files,
                    total_review_edges,
                    total_wi,
                    total_wi_users,
                    total_cycles,
                    total_state,
                    total_cicd,
                    total_deploy,
                    total_incident,
                )

    asyncio.run(_run_with_store(ns.db, db_type, _handler))
    return 0


def main(argv: Optional[List[str]] = None) -> int:
    if os.getenv("DISABLE_DOTENV", "").strip().lower() not in {
        "1",
        "true",
        "yes",
        "on",
    }:
        _load_dotenv(REPO_ROOT / ".env")

    parser = build_parser()
    ns = parser.parse_args(argv)

    level_name = str(getattr(ns, "log_level", "") or "INFO").upper()
    level = getattr(logging, level_name, logging.INFO)
    logging.basicConfig(
        level=level, format="%(asctime)s %(levelname)s %(name)s: %(message)s"
    )

    func = getattr(ns, "func", None)
    if func is None:
        parser.print_help()
        return 2
    return int(func(ns))


if __name__ == "__main__":
    raise SystemExit(main())



"""
GitHub and GitLab connectors for retrieving repository data.

This package provides production-grade connectors for GitHub and GitLab
with automatic pagination, rate limiting, and error handling.
"""

from .base import BatchResult, GitConnector
from .exceptions import (APIException, AuthenticationException,
                         ConnectorException, NotFoundException,
                         PaginationException, RateLimitException)
from .github import GitHubConnector, match_repo_pattern
from .gitlab import GitLabBatchResult, GitLabConnector, match_project_pattern
from .models import (Author, BlameRange, CommitStats, FileBlame, Organization,
                     PullRequest, Repository, RepoStats)

__all__ = [
    # Base class
    "GitConnector",
    # Connectors
    "GitHubConnector",
    "GitLabConnector",
    # GitHub Batch processing
    "BatchResult",
    "match_repo_pattern",
    # GitLab Batch processing
    "GitLabBatchResult",
    "match_project_pattern",
    # Models
    "Organization",
    "Repository",
    "Author",
    "CommitStats",
    "RepoStats",
    "PullRequest",
    "BlameRange",
    "FileBlame",
    # Exceptions
    "ConnectorException",
    "RateLimitException",
    "AuthenticationException",
    "NotFoundException",
    "PaginationException",
    "APIException",
]
